"""
Area managers working with UTM CRS
"""
import logging
from typing import List

import geopandas as gpd
from pydantic import Field

from sentinelhub import Geometry, UtmZoneSplitter

from .base import AreaManager

LOGGER = logging.getLogger(__name__)


class UtmZoneAreaManager(AreaManager):
    """Area manager that splits grid per UTM zones"""

    class Schema(AreaManager.Schema):
        patch_size_x: int = Field(description="A width of each EOPatch in meters")
        patch_size_y: int = Field(description="A height of each EOPatch in meters")
        patch_buffer: float = 0
        offset_x: float = Field(0, description="An offset of tiling grid in horizontal dimension")
        offset_y: float = Field(0, description="An offset of tiling grid in vertical dimension")

    def _get_grid_filename_params(self) -> List[object]:
        """A list of parameters specific to UTM zone splitting"""
        return [
            self.config.patch_size_x,
            self.config.patch_size_y,
            self.config.patch_buffer,
            self.config.offset_x,
            self.config.offset_y,
        ]

    def _create_new_split(self, area_geometry: Geometry) -> List[gpd.GeoDataFrame]:
        """Uses UtmZoneSplitter to create a grid"""
        LOGGER.info("Splitting area geometry into UTM zone grid")
        splitter = UtmZoneSplitter(
            [area_geometry.geometry],
            crs=area_geometry.crs,
            bbox_size=(self.config.patch_size_x, self.config.patch_size_y),
            offset=(self.config.offset_x, self.config.offset_y),
        )
        bbox_list = splitter.get_bbox_list(buffer=self.config.patch_buffer)
        info_list = splitter.get_info_list()

        for tile_info in info_list:
            tile_info["index_n"] = tile_info["index"]

        return self._to_dataframe_grid(bbox_list, info_list, ["index_n", "index_x", "index_y"])



"""
Module containing useful utilities for working with vector data
"""
from typing import List, Union

import pandas as pd
import geopandas as gpd
from geopandas import GeoDataFrame
from shapely.geometry import Polygon, MultiPolygon

from sentinelhub import CRS


def concat_gdf(dataframe_list: List[GeoDataFrame], reproject_crs: Union[CRS, int, None] = None) -> GeoDataFrame:
    """Concatenates together multiple GeoDataFrames, all in the same CRS

    There exists pandas.concat but no geopandas.concat. Therefore this function implements it.

    :param dataframe_list: A list of GeoDataFrames to be concatenated together
    :param reproject_crs: A CRS in which dataframes should be reprojected before being joined
    :return: A joined GeoDataFrame
    """
    if not (dataframe_list and all(isinstance(gdf, GeoDataFrame) for gdf in dataframe_list)):
        raise ValueError(f"Expected a list of GeoDataFrames with at least 1 element, got {dataframe_list}")

    if reproject_crs:
        crs = CRS(reproject_crs).pyproj_crs()
        dataframe_list = [gdf.to_crs(crs) for gdf in dataframe_list]
    else:
        unique_crs = set(vector_gdf.crs for vector_gdf in dataframe_list)
        if len(unique_crs) > 1:
            raise ValueError("GeoDataFrames are in different CRS, therefore reproject_crs parameter should be given")

    return gpd.GeoDataFrame(pd.concat(dataframe_list, ignore_index=True), crs=dataframe_list[0].crs)


def filtered_sjoin(gdf1: GeoDataFrame, gdf2: GeoDataFrame) -> GeoDataFrame:
    """Performs spatial joins of 2 GeoDataFrames and filters repetitions of geometries in the resulting GeoDataFrame.
    Note that geometries in the resulting GeoDataFrame are from the first input GeoDataFrame.

    :param gdf1: First input data frame
    :param gdf2: Second input data frame
    :return: Resulting data frame
    """
    joined_gdf = gpd.sjoin(gdf1, gdf2)
    joined_gdf = joined_gdf.drop(columns=["index_right"])

    unique_index_name = _get_new_unique_column_name(joined_gdf)
    joined_gdf[unique_index_name] = joined_gdf.index.values

    filtered_joined_gdf = joined_gdf.groupby(unique_index_name).first()
    filtered_joined_gdf.index.name = None

    return filtered_joined_gdf


def spatial_difference(gdf1: GeoDataFrame, gdf2: GeoDataFrame) -> GeoDataFrame:
    """Removes polygons from the first GeoDataFrame that intersect with polygons from the second GeoDataFrame

    :param gdf1: First input data frame
    :param gdf2: Second input data frame
    :return: Resulting data frame
    """
    gdf2 = gdf2[["geometry"]]

    intersections = gpd.sjoin(gdf1, gdf2, how="left")
    result_gdf = intersections[intersections["index_right"].isna()]

    result_gdf = result_gdf.drop(columns=["index_right"])
    return result_gdf


def filter_intersecting(gdf: GeoDataFrame, filter_column: str = None) -> GeoDataFrame:
    """Filters out intersecting geometries. A geometry is filtered out if it intersects a geometry with a lower index.

    E.g.: geometries 1 and 2 intersect and geometries 2 and 3 intersect. Then both geometries 2 and 3 will be removed.

    :param gdf: A dataframe to be filtered
    :param filter_column: A name of a column containing indices of geometries, according which to filter. If not
        provided it will use the dataframe index column
    """
    if filter_column is None:
        gdf = gdf.copy(deep=False)
        filter_column = _get_new_unique_column_name(gdf)
        gdf[filter_column] = gdf.index.values

    columns_right = ["geometry", filter_column]
    intersection_gdf = gpd.sjoin(gdf, gdf[columns_right])

    filter_left_column, filter_right_column = f"{filter_column}_left", f"{filter_column}_right"
    valid_column = _get_new_unique_column_name(intersection_gdf)
    intersection_gdf[valid_column] = intersection_gdf[filter_left_column] <= intersection_gdf[filter_right_column]

    intersection_gdf = intersection_gdf.drop(columns=["index_right", filter_left_column, filter_right_column])

    index_column = _get_new_unique_column_name(intersection_gdf)
    intersection_gdf[index_column] = intersection_gdf.index.values

    valid_only_gdf = intersection_gdf.groupby(index_column).filter(lambda df: df[valid_column].all())

    valid_only_gdf = valid_only_gdf.drop(columns=[valid_column])

    filtered_gdf = valid_only_gdf.groupby(index_column).first()
    filtered_gdf.index.name = None

    return filtered_gdf


def _get_new_unique_column_name(df: pd.DataFrame, prefix: str = "temp") -> str:
    """Provides a new column name that doesn't yet exist in a given data frame. The new column will be named in
    a form of <prefix><index>.
    """
    prefix_len = len(prefix)
    suffixes = {column[prefix_len:] for column in df.columns if column.startswith(prefix)}

    index = 1
    while str(index) in suffixes:
        index += 1

    return f"{prefix}{index}"


def count_points(geometry: Union[Polygon, MultiPolygon]) -> int:
    """Counts a number of points for a given geometry, both from exterior and interiors"""
    if isinstance(geometry, Polygon):
        exterior_count = len(geometry.exterior.coords)
        interior_count = sum(len(interior.coords) for interior in geometry.interiors)
        return exterior_count + interior_count

    if isinstance(geometry, MultiPolygon):
        return sum(count_points(subgeometry) for subgeometry in geometry.geoms)

    raise NotImplementedError(f"Counting points for geometry type {type(geometry)} is not yet supported")



"""
Download pipeline that works with Sentinel Hub batch service
"""
import logging
from typing import List, Optional, Tuple, Dict, DefaultDict

from pydantic import Field, conint

from sentinelhub import (
    SentinelHubRequest,
    SentinelHubBatch,
    BatchRequest,
    MimeType,
    BatchRequestStatus,
    BatchTileStatus,
    monitor_batch_job,
    read_data,
    DataCollection,
)

from ..core.pipeline import Pipeline
from ..utils.validators import (
    field_validator,
    optional_field_validator,
    validate_mosaicking_order,
    validate_resampling,
    parse_time_period,
    parse_data_collection,
)
from ..utils.types import Path, TimePeriod

LOGGER = logging.getLogger(__name__)


class BatchDownloadPipeline(Pipeline):
    """Pipeline to start and monitor a Sentinel Hub batch job"""

    class Schema(Pipeline.Schema):
        output_folder_key: str = Field(
            description="Storage manager key pointing to the path where batch results will be saved."
        )
        data_collection: DataCollection = Field(description="Data collection from which data will be downloaded.")
        _validate_data_collection = field_validator("data_collection", parse_data_collection, pre=True)

        time_period: TimePeriod
        _validate_time_period = field_validator("time_period", parse_time_period, pre=True)

        evalscript_path: Path
        tiff_outputs: List[str] = Field(default_factory=list, description="Names of TIFF outputs of a batch job")
        save_userdata: bool = Field(
            False, description="A flag indicating if userdata.json should also be oneof the results of the batch job."
        )

        resampling_type: str = Field(
            "NEAREST", description="A type of downsampling and upsampling used by Sentinel Hub service"
        )
        _validate_resampling_type = field_validator("resampling_type", validate_resampling)

        mosaicking_order: Optional[str] = Field(description="The mosaicking order used by Sentinel Hub service")
        _validate_mosaicking_order = optional_field_validator("mosaicking_order", validate_mosaicking_order)

        monitoring_sleep_time: conint(ge=60) = Field(
            120,
            description=(
                "How many seconds to sleep between two consecutive queries about status of tiles in a batch "
                "job. It should be at least 60 seconds."
            ),
        )

        batch_id: str = Field(
            "",
            description=(
                "An ID of a batch job for this pipeline. If it is given the pipeline will just monitor the "
                "existing batch job. If it is not given it will create a new batch job."
            ),
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.batch_client = SentinelHubBatch(config=self.sh_config)

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Procedure that downloads data to an s3 bucket using batch service"""
        batch_request = self._prepare_and_run_batch_request()
        self.area_manager.batch_id = batch_request.request_id

        LOGGER.info("Monitoring batch job with ID %s", batch_request.request_id)
        results = monitor_batch_job(batch_request, config=self.sh_config, sleep_time=self.config.monitoring_sleep_time)

        if batch_request.value_estimate:
            LOGGER.info("Estimated cost of the batch job: %.1f", batch_request.value_estimate)

        processed = self._get_tile_names_from_results(results, BatchTileStatus.PROCESSED)
        failed = self._get_tile_names_from_results(results, BatchTileStatus.FAILED)
        return processed, failed

    def _prepare_and_run_batch_request(self) -> BatchRequest:
        """Either creates a new job and starts running it or restarts partially failed job or just reports batch
        request status.
        """
        if not self.config.batch_id:
            batch_request = self._create_new_batch_request()
            self.batch_client.start_job(batch_request)
            LOGGER.info("Created a new batch job with ID %s", batch_request.request_id)
            return batch_request

        batch_request = self.batch_client.get_request(self.config.batch_id)
        batch_request.raise_for_status(status=[BatchRequestStatus.FAILED, BatchRequestStatus.CANCELED])

        if batch_request.status is BatchRequestStatus.PARTIAL:
            self.batch_client.restart_job(batch_request)
            LOGGER.info("Restarted batch job %s", batch_request.request_id)
        else:
            LOGGER.info("Current request status is %s", batch_request.status.value)

        return batch_request

    def _create_new_batch_request(self) -> BatchRequest:
        """Defines a new batch request"""
        geometry = self.area_manager.get_area_geometry()

        responses = [
            SentinelHubRequest.output_response(tiff_output, MimeType.TIFF) for tiff_output in self.config.tiff_outputs
        ]
        if self.config.save_userdata:
            responses.append(SentinelHubRequest.output_response("userdata", MimeType.JSON))

        sentinelhub_request = SentinelHubRequest(
            evalscript=read_data(self.config.evalscript_path, data_format=MimeType.TXT),
            input_data=[
                SentinelHubRequest.input_data(
                    data_collection=self.config.data_collection,
                    time_interval=self.config.time_period,
                    upsampling=self.config.resampling_type,
                    downsampling=self.config.resampling_type,
                    mosaicking_order=self.config.mosaicking_order,
                )
            ],
            responses=responses,
            geometry=geometry,
        )

        data_folder = self.storage.get_folder(self.config.output_folder_key, full_path=True).rstrip("/")
        if not self.storage.is_on_aws():
            raise ValueError(f"The data folder path should be on s3 bucket, got {data_folder}")

        return self.batch_client.create(
            sentinelhub_request,
            tiling_grid=SentinelHubBatch.tiling_grid(
                grid_id=self.config.area.tiling_grid_id,
                resolution=self.config.area.resolution,
                buffer=(self.config.area.tile_buffer, self.config.area.tile_buffer),
            ),
            output=SentinelHubBatch.output(default_tile_path=f"{data_folder}/<tileName>/<outputId>.<format>"),
            description=f"eo-grow - {self.__class__.__name__} pipeline with ID {self.pipeline_id}",
        )

    @staticmethod
    def _get_tile_names_from_results(results: DefaultDict[str, List[Dict]], tile_status: BatchTileStatus) -> List[str]:
        """Collects tile names from a dictionary of batch tile results ordered by status"""
        tile_list = results[tile_status.value]
        return [tile["name"] for tile in tile_list]



"""
Module with utilities for creating maps
"""
import logging
import os
import shutil
import subprocess
from tempfile import NamedTemporaryFile
from typing import List, Union

LOGGER = logging.getLogger(__name__)


def cogify_inplace(tiff_file: str, blocksize: int = 2048, nodata: Union[None, int, float] = None):
    """Make the (geotiff) file a cog
    :param tiff_file: .tiff file to cogify
    :param blocksize: block size of tiled COG
    :param nodata: value to be treated as nodata, default value is None
    """
    temp_file = NamedTemporaryFile()
    temp_file.close()

    cogify(tiff_file, temp_file.name, blocksize, nodata=nodata, overwrite=True)
    shutil.move(temp_file.name, tiff_file)


def cogify(
    input_file: str,
    output_file: str,
    blocksize: int = 2048,
    nodata: Union[None, int, float] = None,
    overwrite: bool = False,
):
    """Create a cloud optimized version of input file

    :param input_file: File to cogify
    :param output_file: Resulting cog file
    :param blocksize: block size of tiled COG
    :param nodata: value to be treated as nodata, default value is None
    :param overwrite: If True overwrite the output file if it exists.
    """
    if input_file == output_file:
        raise OSError("Input file is the same as output file.")

    if os.path.exists(output_file):
        if overwrite:
            os.remove(output_file)
        else:
            raise OSError(f"{output_file} exists!")

    gdaladdo_options = f"-r mode --config GDAL_TIFF_OVR_BLOCKSIZE {blocksize} 2 4 8 16 32"

    gdaltranslate_options = (
        "-co TILED=YES -co COPY_SRC_OVERVIEWS=YES "
        f"--config GDAL_TIFF_OVR_BLOCKSIZE {blocksize} -co BLOCKXSIZE={blocksize} "
        f"-co BLOCKYSIZE={blocksize} -co COMPRESS=DEFLATE"
    )

    if nodata is not None:
        gdaltranslate_options += f" -a_nodata {nodata}"

    temp_filename = NamedTemporaryFile()
    temp_filename.close()
    shutil.copyfile(input_file, temp_filename.name)

    LOGGER.info("cogifying %s", input_file)
    subprocess.check_call(f"gdaladdo {temp_filename.name} {gdaladdo_options}", shell=True)
    subprocess.check_call(f"gdal_translate {gdaltranslate_options} {temp_filename.name} {output_file}", shell=True)
    LOGGER.info("cogifying done")


def merge_maps(
    input_filename_list: List[str],
    merged_filename: str,
    *,
    blocksize: int = 2048,
    nodata: Union[None, int, float] = None,
    cogify: bool = False,
    delete_input: bool = False,
):
    """Performs gdal_merge on a set of given geotiff images, make them pixel aligned cogs if `cogify` is True

    :param input_filename_list: A list of input tiff image filenames
    :param merged_filename: Filename of merged tiff image
    :param blocksize: block size of tiled COG
    :param nodata: which values should be treated as nodata in resulting COG, default is None
    :param cogify: If True make the final geotiff a COG.
    :param delete_input: If True input images will be deleted at the end
    """

    merge_tiffs(input_filename_list, merged_filename, overwrite=True, delete_input=delete_input)

    if cogify:
        cogify_inplace(merged_filename, blocksize, nodata=nodata)


def merge_tiffs(
    input_filename_list: List[str], merged_filename: str, *, overwrite: bool = False, delete_input: bool = False
):
    """Performs gdal_merge on a set of given geotiff images

    :param input_filename_list: A list of input tiff image filenames
    :param merged_filename: Filename of merged tiff image
    :param overwrite: If True overwrite the output (merged) file if it exists
    :param delete_input: If True input images will be deleted at the end
    """
    if os.path.exists(merged_filename):
        if overwrite:
            os.remove(merged_filename)
        else:
            raise OSError(f"{merged_filename} exists!")

    LOGGER.info("merging %d tiffs to %s", len(input_filename_list), merged_filename)
    subprocess.check_call(
        ["gdal_merge.py", "-co", "BIGTIFF=YES", "-co", "compress=LZW", "-o", merged_filename, *input_filename_list]
    )
    LOGGER.info("merging done")

    if delete_input:
        LOGGER.info("deleting input files")
        for filename in input_filename_list:
            if os.path.isfile(filename):
                os.remove(filename)



"""
Module implementing export_maps pipelines
"""
import logging
import itertools as it
from typing import List, Optional, Literal, Tuple

import numpy as np
import fs
import fs.copy
from fs.tempfs import TempFS
from pydantic import Field
from eolearn.io import ExportToTiffTask
from eolearn.core import LoadTask, EOWorkflow, FeatureType, linearly_connect_tasks
from eolearn.core.utils.fs import join_path, get_full_path
from eolearn.features import LinearFunctionTask

from ..core.pipeline import Pipeline
from ..utils.map import merge_maps
from ..utils.parallelize import parallelize_with_threads
from ..utils.types import Feature

LOGGER = logging.getLogger(__name__)


class ExportMapsPipeline(Pipeline):
    """Pipeline to export a feature into a tiff map"""

    _merge_map_function = staticmethod(merge_maps)

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(
            description="The storage manager key pointing to the input folder for the export maps pipeline."
        )
        output_folder_key: str = Field(
            description=(
                "The storage manager key pointing to the output folder for the maps in the export maps pipeline."
            )
        )

        feature: Feature
        map_name: str = Field(regex=r".+\.tiff?\b")  # noqa
        map_dtype: Literal["int8", "int16", "uint8", "uint16", "float32"]
        no_data_value: int = Field(0, description="No data value to be passed to GeoTIFFs")
        scale_factor: Optional[float] = Field(description="Feature will be multiplied by this value at export")
        band_indices: Optional[List[int]] = Field(
            description="A list of band indices to be exported for the export feature. Default is all bands"
        )

        cogify: bool = Field(
            False, description="Whether exported GeoTIFFs will be converted into Cloud Optimized GeoTIFFs (COG)"
        )
        force_local_copies: bool = Field(
            False,
            description=(
                "By default copying to local temporary folder will happen only if an AWS S3 path is configured."
                " With this parameter you force to always make copies."
            ),
        )

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Procedure which downloads satellite data"""
        successful, failed = super().run_procedure()

        _, feature_name = self.config.feature
        self.create_maps(feature_name, self.config.map_name, successful)

        return successful, failed

    def build_workflow(self) -> EOWorkflow:
        """Method where workflow is constructed"""
        load_task = LoadTask(
            self.storage.get_folder(self.config.input_folder_key, full_path=True),
            features=[self.config.feature, FeatureType.BBOX],
            config=self.sh_config,
        )
        task_list = [load_task]

        if self.config.scale_factor is not None:
            rescale_task = LinearFunctionTask(self.config.feature, slope=self.config.scale_factor)
            task_list.append(rescale_task)

        feature_name = self.config.feature[1]
        folder = join_path(self.storage.get_folder(self.config.output_folder_key, full_path=True), feature_name)
        export_to_tiff_task = ExportToTiffTask(
            self.config.feature,
            folder=folder,
            no_data_value=self.config.no_data_value,
            image_dtype=np.dtype(self.config.map_dtype),
            band_indices=self.config.band_indices,
            config=self.sh_config,
        )
        task_list.append(export_to_tiff_task)

        return EOWorkflow(linearly_connect_tasks(*task_list))

    def create_maps(self, feature_name: str, map_name: str, patch_names: List[str], utm_split: bool = True) -> None:
        """A method which creates a joined GeoTIFF maps from a list of exported GeoTIFFs"""
        if not patch_names:
            raise ValueError("Cannot create map with an empty list of EOPatch names")

        folder = fs.path.join(self.storage.get_folder(self.config.output_folder_key), feature_name)

        if utm_split:  # TODO: this can be parallelized but _create_single_map would have to become static
            crs_eopatch_dict = self.eopatch_manager.split_by_utm(patch_names)

            for crs, eopatch_list in crs_eopatch_dict.items():
                utm_map_name = f"utm{crs.epsg}_{map_name}"
                self._create_single_map(folder, eopatch_list, utm_map_name)
        else:
            self._create_single_map(folder, patch_names, map_name)

    def _create_single_map(self, folder: str, eopatch_list: List[str], map_name: str) -> None:
        """Creates a single map"""
        make_local_copies = self.storage.is_on_aws() or self.config.force_local_copies

        geotiff_names = [self.get_geotiff_name(name) for name in eopatch_list]
        geotiff_paths = [fs.path.join(folder, name) for name in geotiff_names]
        merged_map_path = fs.path.join(folder, map_name)

        if make_local_copies:
            temp_fs = TempFS(identifier="_merge_maps_temp")

            LOGGER.info("Copying tiffs to a temporary local folder %s", temp_fs.getsyspath("/"))
            parallelize_with_threads(
                fs.copy.copy_file, it.repeat(self.storage.filesystem), geotiff_paths, it.repeat(temp_fs), geotiff_names
            )

            sys_paths = [temp_fs.getsyspath(name) for name in geotiff_names]
            sys_map_path = temp_fs.getsyspath(map_name)
        else:
            sys_paths = [self.storage.filesystem.getsyspath(path) for path in geotiff_paths]
            sys_map_path = self.storage.filesystem.getsyspath(merged_map_path)

        self._merge_map_function(sys_paths, sys_map_path, cogify=self.config.cogify, delete_input=False)

        if make_local_copies:
            fs.copy.copy_file(temp_fs, map_name, self.storage.filesystem, merged_map_path)
            temp_fs.close()
        LOGGER.info("Merged map is saved at %s", get_full_path(self.storage.filesystem, merged_map_path))

        parallelize_with_threads(self.storage.filesystem.remove, geotiff_paths)
        LOGGER.info("Cleaned original temporary maps")

    def get_geotiff_name(self, name: str) -> str:
        """Creates a unique name of a geotiff image"""
        return f"{name}_{self.__class__.__name__}_{self.pipeline_id}.tiff"

    def get_execution_arguments(self, workflow: EOWorkflow) -> List[dict]:
        exec_args = super().get_execution_arguments(workflow)
        nodes = workflow.get_nodes()
        for name, single_exec_dict in zip(self.patch_list, exec_args):
            for node in nodes:
                if isinstance(node.task, ExportToTiffTask):
                    single_exec_dict[node] = dict(filename=self.get_geotiff_name(name))

        return exec_args



"""
Module implementing sampling pipelines
"""
import abc
from typing import Optional, Dict, List, Tuple, Union

import numpy as np
from pydantic import Field, validator

from eolearn.core import (
    FeatureType,
    SaveTask,
    LoadTask,
    OverwritePermission,
    EOWorkflow,
    MergeEOPatchesTask,
    EONode,
)
from eolearn.ml_tools import FractionSamplingTask, BlockSamplingTask, GridSamplingTask
from eolearn.geometry import MorphologicalStructFactory, MorphologicalOperations

from ..core.config import Config
from ..core.pipeline import Pipeline
from ..tasks.common import ClassFilterTask
from ..utils.filter import get_patches_without_all_features
from ..utils.types import Feature


class BaseSamplingPipeline(Pipeline, metaclass=abc.ABCMeta):
    """Pipeline to run sampling on EOPatches"""

    class Schema(Pipeline.Schema):
        output_folder_key: str = Field(
            description="The storage manager key pointing to the output folder for the Sampline pipeline."
        )
        apply_to: Dict[str, Dict[str, List[str]]] = Field(
            description=(
                "A dictionary defining which features to sample, its structure is "
                "{folder_key: {feature_type: [feature_name]}}"
            ),
        )
        mask_of_samples_name: Optional[str] = Field(
            description=(
                "A name of a mask timeless output feature with information which pixels were sampled and how many times"
            )
        )
        sampled_suffix: Optional[str] = Field(
            description=(
                "If provided features are saved with a suffix, e.g. for suffix SAMPLED the sampled FEATURES are "
                "saved as FEATURES_SAMPLED."
            )
        )
        compress_level: int = Field(1, description="Level of compression used in saving eopatches")

    def filter_patch_list(self, patch_list: List[str]) -> List[str]:
        """Filter output EOPatches that have already been processed"""
        filtered_patch_list = get_patches_without_all_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            self._get_output_features(),
        )

        return filtered_patch_list

    def build_workflow(self) -> EOWorkflow:
        """Creates workflow that is divided into the following sub-parts:

        1. loading data,
        2. preprocessing steps,
        3. sampling features
        4. saving results
        """
        loading_node = self._get_loading_node()
        preprocessing_node = self._get_preprocessing_node(loading_node)
        sampling_node = self._get_sampling_node(preprocessing_node)

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key, full_path=True),
            features=self._get_output_features(),
            compress_level=self.config.compress_level,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            config=self.sh_config,
        )

        return EOWorkflow.from_endnodes(EONode(save_task, inputs=[sampling_node]))

    def _get_loading_node(self) -> EONode:
        """Prepares nodes for loading and joining EOPatches."""
        load_nodes = []

        for folder_name, features in self.config.apply_to.items():
            load_features = []

            for feature_type, feature_names in features.items():
                feature_type = FeatureType(feature_type)

                if not feature_type.is_spatial():
                    raise ValueError(f"Only spatial features can be sampled, but found {feature_type}: {feature_names}")

                for feature_name in feature_names:
                    load_features.append((feature_type, feature_name))

            load_features.append(FeatureType.BBOX)
            if any(FeatureType(feature_type).is_temporal() for feature_type in features):
                load_features.append(FeatureType.TIMESTAMP)

            load_task = LoadTask(
                self.storage.get_folder(folder_name, full_path=True),
                lazy_loading=True,
                features=load_features,
                config=self.sh_config,
            )
            load_nodes.append(EONode(load_task, name=f"Load from {folder_name}"))

        return EONode(MergeEOPatchesTask(), inputs=load_nodes)

    def _get_preprocessing_node(self, previous_node: EONode) -> EONode:  # pylint: disable=no-self-use
        """The default implementation doesn't add any preprocessing steps"""
        return previous_node

    @abc.abstractmethod
    def _get_sampling_node(self, previous_node: EONode) -> EONode:
        """Method to prepare sampling nodes"""

    def _get_features_to_sample(self) -> List[Tuple[FeatureType, str, str]]:
        """Get a list of features that will be sampled, together with their new names"""
        features_to_sample = []
        for _, features in self.config.apply_to.items():
            for feature_type, feature_names in features.items():
                feature_type = FeatureType(feature_type)

                for feature_name in feature_names:
                    if self.config.sampled_suffix is None:
                        features_to_sample.append((feature_type, feature_name, feature_name))
                    else:
                        features_to_sample.append(
                            (feature_type, feature_name, f"{feature_name}_{self.config.sampled_suffix}")
                        )

        return features_to_sample

    def _get_mask_of_samples_feature(self) -> Optional[Feature]:
        """Provide a mask of samples feature"""
        if self.config.mask_of_samples_name:
            return FeatureType.MASK_TIMELESS, self.config.mask_of_samples_name
        return None

    def _get_output_features(self) -> List[Feature]:
        """Get a list of features that will be saved as an output of the pipeline"""
        output_features = [FeatureType.BBOX]
        features_to_sample = self._get_features_to_sample()

        for feature_type, _, sampled_feature_name in features_to_sample:
            output_features.append((feature_type, sampled_feature_name))

        if self.config.mask_of_samples_name:
            output_features.append(self._get_mask_of_samples_feature())

        if any(feature_type.is_temporal() for feature_type, _, _ in features_to_sample):
            output_features.append(FeatureType.TIMESTAMP)
        return output_features


class BaseRandomSamplingPipeline(BaseSamplingPipeline, metaclass=abc.ABCMeta):
    """A base class for all sampling pipeline that work on random selection of samples"""

    class Schema(BaseSamplingPipeline.Schema):
        seed: Optional[int] = Field(
            42, description="A random generator seed to be used in order to obtain the same results every pipeline run."
        )

    def __init__(self, config: Config):
        super().__init__(config)
        self._sampling_node_uid = None

    def get_execution_arguments(self, workflow: EOWorkflow) -> List[Dict[str, object]]:
        """Extends the basic method for adding execution arguments by adding seed arguments a sampling task"""
        exec_args = super().get_execution_arguments(workflow)

        sampling_node = workflow.get_node_with_uid(self._sampling_node_uid)
        if sampling_node is None:
            return exec_args

        generator = np.random.default_rng(seed=self.config.seed)

        for arg_index in range(len(self.patch_list)):
            exec_args[arg_index][sampling_node] = dict(seed=generator.integers(low=0, high=2**32))

        return exec_args


class FractionSamplingPipeline(BaseRandomSamplingPipeline):
    """A pipeline to sample per-class with different distributions"""

    class Schema(BaseRandomSamplingPipeline.Schema):
        sampling_feature_name: str = Field(
            description="Name of MASK_TIMELESS feature to be used to create sample point"
        )
        erosion_dict: Optional[Dict[float, List[int]]] = Field(
            description="A dictionary specifying disc radius of erosion operation to be applied to a list of label IDs",
            example={2.5: [1, 3, 4], 1.0: [2]},
        )
        fraction_of_samples: Union[float, Dict[int, float]] = Field(
            description=(
                "A fraction or a dictionary of per-class fractions of valid pixels to sample from the the sampling "
                "feature."
            )
        )
        exclude_values: List[int] = Field(default_factory=list, description="Values to be excluded from sampling")

    def _get_preprocessing_node(self, previous_node: EONode) -> EONode:
        """Preprocessing that applies erosion on sampling feature values"""
        if self.config.erosion_dict is None:
            return previous_node

        end_node = previous_node
        for radius, labels in self.config.erosion_dict.items():
            task = ClassFilterTask(
                (FeatureType.MASK_TIMELESS, self.config.sampling_feature_name),
                labels,
                MorphologicalOperations.EROSION,
                struct_elem=MorphologicalStructFactory.get_disk(radius),
            )
            end_node = EONode(task, inputs=[end_node])

        return end_node

    def _get_sampling_node(self, previous_node: EONode) -> EONode:
        """Prepare the sampling task"""
        task = FractionSamplingTask(
            features_to_sample=self._get_features_to_sample(),
            sampling_feature=(FeatureType.MASK_TIMELESS, self.config.sampling_feature_name),
            fraction=self.config.fraction_of_samples,
            exclude_values=self.config.exclude_values,
            mask_of_samples=self._get_mask_of_samples_feature(),
        )
        node = EONode(task, inputs=[previous_node])
        self._sampling_node_uid = node.uid

        return node


class BlockSamplingPipeline(BaseRandomSamplingPipeline):
    """A pipeline to randomly sample blocks"""

    class Schema(BaseRandomSamplingPipeline.Schema):
        sample_size: Tuple[int, int] = Field(description="A height and width of each block in pixels.")

        number_of_samples: Optional[int] = Field(
            description=(
                "A number of samples to be sampled. Exactly one of parameters fraction_of_samples and number_of_samples"
                " has to be given."
            )
        )
        fraction_of_samples: Optional[float] = Field(
            description=(
                "A percentage of samples to be sampled. Exactly one of parameters fraction_of_samples and "
                "number_of_samples has to be given."
            )
        )

        @validator("fraction_of_samples", always=True)
        def validate_sampling_params(cls, fraction, values):  # pylint: disable=no-self-use, no-self-argument
            """Makes sure only one of the sampling parameters has been given"""
            assert (fraction is None) != (
                values.get("number_of_samples") is None
            ), "Exactly one of the parameters fraction_of_samples and number_of_samples should be defined"
            return fraction

    def _get_sampling_node(self, previous_node: EONode) -> EONode:
        """Prepare the sampling task"""
        task = BlockSamplingTask(
            features_to_sample=self._get_features_to_sample(),
            amount=self.config.fraction_of_samples or self.config.number_of_samples,
            sample_size=self.config.sample_size,
            mask_of_samples=self._get_mask_of_samples_feature(),
        )

        node = EONode(task, inputs=[previous_node])
        self._sampling_node_uid = node.uid

        return node


class GridSamplingPipeline(BaseSamplingPipeline):
    """A pipeline to sample blocks in a regular grid"""

    class Schema(BaseSamplingPipeline.Schema):
        sample_size: Tuple[int, int] = Field(description="A height and width of each block in pixels.")
        stride: Tuple[int, int] = Field(
            description=(
                "A tuple describing a distance between upper left corners of two consecutive sampled blocks. "
                "The first number is the vertical distance and the second number the horizontal distance. If stride in "
                "smaller than sample_size in any dimensions then sampled blocks will overlap."
            )
        )

    def _get_sampling_node(self, previous_node: EONode) -> EONode:
        """Prepare the sampling task"""
        task = GridSamplingTask(
            features_to_sample=self._get_features_to_sample(),
            sample_size=self.config.sample_size,
            stride=self.config.stride,
            mask_of_samples=self._get_mask_of_samples_feature(),
        )

        return EONode(task, inputs=[previous_node])



"""
Module implementing management of configurations
"""
import base64
import binascii
import copy
import json
import pprint
import os
import re
from typing import Optional, Set, Dict, Callable, Tuple, Sequence, Union

import rapidjson
import fs.path
from munch import Munch, munchify, unmunchify
from eolearn.core.utils.fs import get_base_filesystem_and_path, join_path, get_full_path

from ..utils.general import jsonify
from ..utils.meta import get_os_import_path


class _BaseConfig:
    """A base class for configuration objects implementing methods that are common between `Config` and `ConfigList`"""

    @staticmethod
    def from_path(path: str) -> Union["Config", "ConfigList"]:
        """Loads from path in applies operations defined in `interpret_config_from_path`"""
        return interpret_config_from_path(path)

    @staticmethod
    def from_string(config_str: str) -> Union["Config", "ConfigList"]:
        """Creates a config object either from a base64-encoded string or a file path"""
        return get_config_from_string(config_str)

    def encode(self) -> str:
        """Dumps config into a json and the encodes it with base64

        :return: A base64-encoded string
        """
        json_string = json.dumps(self, default=jsonify)
        return base64.b64encode(json_string.encode()).decode()


class Config(Munch, _BaseConfig):
    """Class for handling `eo-grow` configurations

    This is a wrapper around `munch.Munch` and it allows to use dictionary parameters as attributes. E.g.
    besides `config['param']` you can also call `config.param`.

    Note that `Config(config_dict)` will only convert the top dictionary level into `Config`. Because of that the
    preferred way of loading is `Config.from_dict(config_dict)`.
    """

    @classmethod
    def from_dict(cls, config_dict: dict) -> "Config":
        """Converts a normal dictionary into a `Config` object.

        The conversion is applied recursively on all nested dictionaries.
        """
        if not isinstance(config_dict, dict):
            raise ValueError(f"This method expected a dictionary, got {type(config_dict)}")
        return munchify(config_dict, factory=cls)

    def __repr__(self) -> str:
        """A nicer representation with pprint of the dictionary"""
        return f"{self.__class__.__name__}({{\n {pprint.pformat(dict(self))[1: -1]}\n}})"

    def __getattr__(self, item: str):
        """This additionally loads values of environmental variables

        This method performs the 3rd stage of language interpretation as described in
        `eo-grow/documentation/config-language.md`.
        """
        value = super().__getattr__(item)

        if isinstance(value, str):
            return _resolve_env_variables(value)

        if isinstance(value, list):
            return [(_resolve_env_variables(param) if isinstance(param, str) else param) for param in value]
        return value

    def add_cli_variables(self, cli_variables: Sequence[str]):
        """Adds variables passed in the CLI to the config. Has to be used before interpretation!"""
        cli_variable_mapping = dict(_parse_cli_variable(cli_var) for cli_var in cli_variables)
        current_variables = self.get("variables", {})
        current_variables.update(cli_variable_mapping)
        self["variables"] = current_variables

    def interpret(self) -> "Config":
        """Applies operations defined in `interpret_config_from_dict`"""
        return interpret_config_from_dict(self)

    def to_dict(self) -> dict:
        """Converts the object into a normal dictionary"""
        return unmunchify(self)


class ConfigList(list, _BaseConfig):
    """A class for handling a list of configurations"""

    @classmethod
    def from_list(cls, config_list: Sequence[dict]) -> "ConfigList":
        """Converts a normal list of dictionaries into a `ConfigList` object."""
        return cls(Config.from_dict(config) for config in config_list)

    def __repr__(self) -> str:
        """A nicer representation with pprint of the list"""
        return f"{self.__class__.__name__}([\n {pprint.pformat(list(self))[1: -1]}\n])"


def get_config_from_string(config_str: str) -> Union[Config, ConfigList]:
    """Provides a config object by either decoding a base64-encoded string or load it from a file path."""
    try:
        decoded_string = base64.b64decode(config_str.encode()).decode()
        raw_config = json.loads(decoded_string)

        if isinstance(raw_config, dict):
            return Config.from_dict(raw_config)
        return ConfigList.from_list(raw_config)
    except (binascii.Error, UnicodeDecodeError, json.JSONDecodeError):
        pass

    return interpret_config_from_path(config_str)


def interpret_config_from_path(path: str, used_config_paths: Optional[Set[str]] = None) -> Union[Config, ConfigList]:
    """Loads and builds a config from parameters stored in files

    This function performs the 1st stage of language interpretation as described in
    eo-grow/documentation/config-language.md`.

    :param path: A full path where config is located
    :param used_config_paths: Just a helper parameter to prevent cyclic config imports.
    :return: A fully built config object.
    """
    filesystem, path = get_base_filesystem_and_path(path)
    with filesystem.open(path, "r") as file_handle:
        config = rapidjson.load(file_handle, parse_mode=rapidjson.PM_COMMENTS)

    used_config_paths = copy.copy(used_config_paths) or set()
    used_config_paths.add(get_full_path(filesystem, path))

    full_folder_path = get_full_path(filesystem, fs.path.dirname(path))
    config = _recursive_apply_to_strings(config, lambda value: _resolve_config_paths(value, full_folder_path))

    config = _recursive_config_build(config, used_config_paths)

    if isinstance(config, dict):
        return Config.from_dict(config)
    return ConfigList.from_list(config)


def _recursive_config_build(config, used_config_paths: Set[str]):
    """Recursively builds a configuration object by replacing dictionary items in form of

        `'**key': 'file path to another config'`

    with a content of a configuration file, referenced with the path. The items from both configuration objects are
    joined recursively.
    """
    if isinstance(config, dict):
        joint_config = {}
        imported_configs = []

        for key, value in config.items():
            if not isinstance(key, str):
                raise ValueError(f"Dictionary keys should always be strings, but found: {key}")

            if key.startswith("**"):
                if value in used_config_paths:
                    raise ValueError("Detected a cyclic import of configs")

                imported_config = interpret_config_from_path(value, used_config_paths=used_config_paths)
                imported_configs.append(imported_config)
            else:
                joint_config[key] = _recursive_config_build(value, used_config_paths)

        for imported_config in imported_configs:
            joint_config = recursive_config_join(joint_config, imported_config)

        return joint_config

    if isinstance(config, list):
        return [_recursive_config_build(value, used_config_paths) for value in config]

    return config


def interpret_config_from_dict(config: dict) -> Config:
    """Applies config language rules to a loaded config

    This function performs the 2nd stage of language interpretation as described in
    `eo-grow/documentation/config-language.md`.
    """
    _recursive_check_config(config)
    config = _recursive_apply_to_strings(config, _resolve_import_paths)

    variable_mapping = config.pop("variables", {})

    for variable_name in variable_mapping:
        if not re.fullmatch(r"\w+", variable_name):
            raise ValueError(f"Variable name {variable_name} contains illegal characters")

    config = _recursive_apply_to_strings(config, lambda config_str: _resolve_variables(config_str, variable_mapping))

    return Config.from_dict(config)


def _resolve_config_paths(config_str: str, config_path: str) -> str:
    """Replaces `${config_path}` with an actual path"""
    new_config_str = re.sub(r"\${config_path}", config_path, config_str)

    if new_config_str != config_str:
        return join_path(new_config_str)
    return new_config_str


def _resolve_import_paths(config_str: str) -> str:
    """Replaces `${import_path:package.module}` with an actual local path"""
    new_config_str = re.sub(r"\${import_path:([\w.]+)}", _sub_import_path, config_str)

    if new_config_str != config_str:
        return join_path(new_config_str)
    return new_config_str


def _sub_import_path(match: re.Match) -> str:
    """Substitutes a regex match with a local file path to a module or a package"""
    import_module = match.group(1)
    return os.path.dirname(get_os_import_path(import_module))


def _resolve_variables(config_str: str, variable_mapping: Dict[str, str]) -> str:
    """Replaces `${var:variable_name}` with a replacement value defined under variables"""
    return re.sub(r"\${var:(\w+)}", lambda match: _sub_variable(match, variable_mapping), config_str)


def _sub_variable(match: re.Match, variable_mapping: Dict[str, str]) -> str:
    """Substitutes a regex match with a new variable according to the mapping"""
    variable_name = match.group(1)
    if variable_name in variable_mapping:
        return str(variable_mapping[variable_name])
    raise ValueError(f"Variable name '{variable_name}' doesn't exist in a config dictionary of variables.")


def _resolve_env_variables(config_str: str) -> str:
    """Replaces `${env:variable_name}` with an environmental variable"""
    return re.sub(r"\${env:(\w+)}", _sub_env_variable, config_str)


def _sub_env_variable(match: re.Match) -> str:
    """Substitutes a regex match with a value set under environmental variable"""
    env_variable_name = match.group(1)
    env_var_value = os.getenv(env_variable_name)
    if env_var_value is None:
        raise KeyError(f"Environmental variable {env_variable_name} is not present in environment")
    return env_var_value


def _recursive_apply_to_strings(config, function: Callable):
    """Recursively applies a function on all string values (and not keys) of a nested config object"""
    if isinstance(config, dict):
        return {key: _recursive_apply_to_strings(value, function) for key, value in config.items()}

    if isinstance(config, list):
        return [_recursive_apply_to_strings(value, function) for value in config]

    if isinstance(config, str):
        return function(config)
    return config


def _recursive_check_config(config: dict):
    """Recursively checks if a config object conforms to some basic rules

    :raises: ValueError
    """
    if isinstance(config, dict):
        for key, value in config.items():
            if not isinstance(key, str):
                raise ValueError(f"Config keys should be strings but {key} found")
            _recursive_check_config(value)

    elif isinstance(config, list):
        for value in config:
            _recursive_check_config(value)


def recursive_config_join(config1: dict, config2: dict) -> dict:
    """Recursively join 2 config objects, where config1 values override config2 values"""
    for key, value in config2.items():
        if key not in config1:
            config1[key] = value
        elif isinstance(config1[key], dict) and isinstance(value, dict):
            config1[key] = recursive_config_join(config1[key], value)

    return config1


def _parse_cli_variable(mapping_str: str) -> Tuple[str, str]:
    """Checks that the input is of shape `name:value` and then splits it into a tuple"""
    match = re.match(r"(?P<name>.+?):(?P<value>.+)", mapping_str)
    if match is None:
        raise ValueError(f'CLI variable input {mapping_str} is not of form `"name:value"`')
    parsed = match.groupdict()
    return parsed["name"], parsed["value"]



from typing import Type
from pydantic import BaseModel

from .config import Config
from ..utils.meta import collect_schema


class EOGrowObject:
    """A base object object in `eo-grow` framework"""

    def __init__(self, config: dict):
        config = config if isinstance(config, Config) else Config.from_dict(config)

        self.schema = self._initialize_schema()
        self._config = self._prepare_config(config)

    def _initialize_schema(self) -> Type[BaseModel]:
        schema_object = collect_schema(self)
        return schema_object

    def _prepare_config(self, config: Config) -> Config:
        """Interprets and validates configuration dictionary"""
        config = config.interpret()
        parsed_config = self.schema.parse_obj(config)

        return Config.from_dict(parsed_config.dict())

    @property
    def config(self) -> Config:
        """A public property that provides object's configuration"""
        return self._config



"""
Area managers for Sentinel Hub batch grids
"""
from typing import List

from geopandas.geodataframe import GeoDataFrame
from pydantic import Field

from sentinelhub import BBox, SentinelHubBatch, BatchSplitter, BatchRequestStatus
from sentinelhub.sentinelhub_batch import BatchRequest

from .base import AreaManager


class BatchAreaManager(AreaManager):
    """Area manager that is based on Sentinel Hub Batch service tiling"""

    class Schema(AreaManager.Schema):
        """A schema used for pipelines based on batch tiling"""

        tiling_grid_id: int
        resolution: float
        tile_buffer: int = Field(0, description="Number of pixels for which each tile will be buffered")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.batch_id = ""

    def _get_grid_filename_params(self) -> List[object]:
        """Puts together batch parameters"""
        return [self.config.tiling_grid_id, self.config.resolution, self.config.tile_buffer]

    def _create_new_split(self, *_, **__) -> List[GeoDataFrame]:
        """Instead of creating a split it loads tiles from the service"""
        if not self.batch_id:
            raise ValueError(
                "Trying to create a new batch grid but cannot collect tile geometries because 'batch_id' has not been "
                "given. Make sure that you are running a pipeline that creates a new batch job."
            )

        batch_request = SentinelHubBatch().get_request(self.batch_id)
        self._verify_batch_request(batch_request)

        splitter = BatchSplitter(batch_request=batch_request)
        bbox_list = splitter.get_bbox_list()
        info_list = splitter.get_info_list()

        if not all(bbox.crs.is_utm() for bbox in bbox_list):
            raise NotImplementedError("So far we only support UTM-based batch tiling grids")

        # The problem is that Sentinel Hub service returns unbuffered tile geometries in WGS84. BatchSplitter then
        # transforms them into UTM, but here we still have to fix numerical errors by rounding and then apply buffer.
        bbox_list = [self._fix_batch_bbox(bbox) for bbox in bbox_list]

        # Sentinel Hub service provides tile info in an arbitrary order. We sort by tile names for consistency.
        bbox_list, info_list = zip(*sorted(zip(bbox_list, info_list), key=lambda pair: pair[1]["name"]))
        for index, info in enumerate(info_list):
            info["index_n"] = index

        return self._to_dataframe_grid(bbox_list, info_list, ["index_n", "id", "name", "cost"])

    def _verify_batch_request(self, batch_request: BatchRequest):
        """Verifies that given batch request has finished and that it has the same tiling grid parameters as
        they are written in the config.
        """
        batch_request.raise_for_status(status=[BatchRequestStatus.FAILED, BatchRequestStatus.CANCELED])

        expected_tiling_grid_params = {
            "id": self.config.tiling_grid_id,
            "bufferY": self.config.tile_buffer,
            "bufferX": self.config.tile_buffer,
            "resolution": self.config.resolution,
        }
        if batch_request.tiling_grid != expected_tiling_grid_params:
            raise ValueError(
                f"Tiling grid parameters in config are {expected_tiling_grid_params} but given batch "
                f"request has parameters {batch_request.tiling_grid}"
            )

    def _fix_batch_bbox(self, bbox: BBox) -> BBox:
        """Fixes a batch tile bounding box so that it will be the same as in produced tiles on the bucket"""
        min_x, min_y, max_x, max_y = list(map(round, bbox))

        buffer_meters = self.config.tile_buffer * self.config.resolution

        min_x, min_y = min_x - buffer_meters, min_y - buffer_meters
        max_x, max_y = max_x + buffer_meters, max_y + buffer_meters

        return BBox((min_x, min_y, max_x, max_y), crs=bbox.crs)



"""
Module with utilities for logging
"""
import json
import logging
import sys
import time
from logging import Handler, StreamHandler, FileHandler, Formatter, Filter, LogRecord
from typing import Optional, List, Union, Sequence

import fs
from fs.errors import FilesystemClosed
from pydantic import Field
from sentinelhub import SHConfig
from eolearn.core.utils.fs import join_path

from .base import EOGrowObject
from .config import Config
from .schemas import ManagerSchema
from .storage import StorageManager
from ..utils.fs import LocalFile
from ..utils.general import jsonify
from ..utils.logging import get_instance_info
from ..utils.meta import get_package_versions


class LoggingManager(EOGrowObject):
    """A class that manages logging specifics"""

    class Schema(ManagerSchema):
        save_logs: bool = Field(
            False,
            description=(
                "A flag to determine if pipeline logs and reports will be saved to "
                "logs folder. This includes potential EOExecution reports and logs."
            ),
        )
        pipeline_ignore_packages: Optional[List[str]] = Field(
            description=(
                "Names of packages which logs will not be written to the main pipeline log file. The default null "
                "value means that a default list of packages will be used."
            )
        )
        pipeline_logs_backup_interval: float = Field(
            60,
            description=(
                "When working with a remote storage this parameter defines a minimal number of seconds between "
                "two consecutive times when pipeline log file will be copied into the remote storage."
            ),
        )
        eoexecution_ignore_packages: Optional[List[str]] = Field(
            description=(
                "Names of packages which logs will not be written to EOExecution log files. The default null value "
                "means that a default list of packages will be used."
            )
        )
        include_logs_to_report: bool = Field(
            False,
            description=(
                "If log files should be parsed into an EOExecution report file or just linked. When working "
                "with larger number of EOPatches the recommended option is False."
            ),
        )

        show_logs: bool = Field(False, description="Shows basic pipeline execution logs at stdout.")
        stdout_log_packages: Optional[List[str]] = Field(
            description=(
                "Names of packages which logs will be written to stdout. The default null value means that a default "
                "list of packages will be used."
            )
        )

        capture_warnings: bool = Field(
            True,
            description=(
                "If warnings should be treated as logs and with save_logs=True written into log files instead of "
                "being printed in stderr."
            ),
        )

    def __init__(self, config: Config, storage: StorageManager):
        """
        :param config: A configuration file
        :param storage: An instance of StorageManager class
        """
        super().__init__(config)

        self.storage = storage

    def get_pipeline_logs_folder(self, pipeline_execution_name: str, full_path: bool = False) -> str:
        """Provides path to the folder where logs of this pipeline execution will be saved

        :param pipeline_execution_name: Name of current pipeline execution
        :param full_path: If it should provide a full absolute path or a path relative to the filesystem object
        """
        main_logs_folder = self.storage.get_logs_folder(full_path=full_path)
        if full_path:
            return join_path(main_logs_folder, pipeline_execution_name)
        return fs.path.combine(main_logs_folder, pipeline_execution_name)

    def start_logging(self, pipeline_execution_name: str) -> List[Handler]:
        """Creates a folder for logs and sets up (and returns) logging handlers

        Supported handlers:
        - Writing to a file in pipeline logs folder
        - Printing logs to a standard output
        """
        if self.config.save_logs:
            logs_folder = self.get_pipeline_logs_folder(pipeline_execution_name)
            self.storage.filesystem.makedirs(logs_folder, recreate=True)

        global_logger = logging.getLogger()
        global_logger.setLevel(logging.DEBUG)

        for default_handler in global_logger.handlers:
            default_handler.setLevel(logging.WARNING)

        new_handlers: List[Handler] = []

        if self.config.save_logs:
            file_handler = self._create_file_handler(pipeline_execution_name)
            global_logger.addHandler(file_handler)
            new_handlers.append(file_handler)

        if self.config.show_logs:
            stdout_handler = self._create_stdout_handler()
            global_logger.addHandler(stdout_handler)
            new_handlers.append(stdout_handler)

        if self.config.capture_warnings:
            logging.captureWarnings(True)

        return new_handlers

    def _create_file_handler(self, pipeline_execution_name: str) -> Handler:
        """Creates a logging handler to write a pipeline log to a file."""
        logs_filename = fs.path.combine(self.get_pipeline_logs_folder(pipeline_execution_name), "pipeline.log")
        file_handler = RegularBackupHandler(
            logs_filename,
            filesystem=self.storage.filesystem,
            backup_interval=self.config.pipeline_logs_backup_interval,
            encoding="utf-8",
        )

        formatter = Formatter(
            "%(levelname)s %(asctime)s %(name)s:%(lineno)d:\n\t%(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        file_handler.setFormatter(formatter)

        file_handler.addFilter(LogFileFilter())

        return file_handler

    def _create_stdout_handler(self) -> Handler:
        """Creates a logging handler to write logs into a standard output."""
        stdout_handler = StreamHandler(sys.stdout)
        stdout_handler.setLevel(logging.INFO)

        formatter = Formatter("%(levelname)s %(name)s:%(lineno)d: %(message)s")
        stdout_handler.setFormatter(formatter)

        stdout_handler.addFilter(StdoutFilter(log_packages=self.config.stdout_log_packages))

        return stdout_handler

    def stop_logging(self, handlers: List[Handler]):
        """Updates logs, removes pipeline handlers from the global logger and puts global logging level back to
        default
        """
        if self.config.capture_warnings:
            logging.captureWarnings(False)

        global_logger = logging.getLogger()
        for handler in handlers:
            handler.close()
            global_logger.removeHandler(handler)

        global_logger.setLevel(logging.WARNING)

    def update_pipeline_report(
        self,
        pipeline_execution_name: str,
        pipeline_config: Config,
        pipeline_id: str,
        pipeline_timestamp: str,
        elapsed_time: Optional[float] = None,
    ):
        """A method in charge of preparing a report about pipeline run.

        Content of a report:
         - pipeline configuration parameters,
         - pipeline execution stats,
         - versions of Python and Python packages,
         - information about a compute instance on which the pipeline is running.
        """
        if not self.config.save_logs:
            return

        report = {
            "config_parameters": pipeline_config,
            "pipeline_execution_stats": {
                "pipeline_id": pipeline_id,
                "start_time": pipeline_timestamp,
                "elapsed_time": "<Not yet finished>" if elapsed_time is None else elapsed_time,
            },
            "versions": {"Python": sys.version, **get_package_versions()},
            "instance_info": get_instance_info(),
        }

        report_filename = fs.path.combine(
            self.get_pipeline_logs_folder(pipeline_execution_name), "pipeline-report.json"
        )
        with self.storage.filesystem.open(report_filename, "w") as report_file:
            json.dump(report, report_file, indent=2, default=jsonify)

    def save_eopatch_execution_status(self, pipeline_execution_name: str, finished: list, failed: list):
        """Saves lists of EOPatch names for which execution either finished successfully or failed"""
        if not self.config.save_logs:
            return

        filesystem = self.storage.filesystem
        logs_folder = self.get_pipeline_logs_folder(pipeline_execution_name)

        for eopatches, filename in [(finished, "finished.json"), (failed, "failed.json")]:
            path = fs.path.combine(logs_folder, filename)
            with filesystem.open(path, "w") as file:
                json.dump(eopatches, file, indent=2)


class FilesystemHandler(FileHandler):
    """A filesystem abstraction of FileHandler

    In case the handler gets a local path it behaves the same as FileHandler. In case it gets a remote path it writes
    logs first to a local path and then copies them to the remote location.
    """

    def __init__(self, path: str, filesystem: Optional[fs.base.FS] = None, config: Optional[SHConfig] = None, **kwargs):
        """
        :param path: A path to a log file. It should be an absolute path if filesystem object is not given and relative
            otherwise.
        :param filesystem: A filesystem to where logs will be written.
        :param config: A config object holding credentials.
        :param kwargs: Keyword arguments that will be propagated to FileHandler.
        """
        self.local_file = LocalFile(path, mode="w", filesystem=filesystem, config=config)

        super().__init__(self.local_file.path, **kwargs)

    def close(self) -> None:
        """Closes logging and closes the local file"""
        super().close()
        try:
            self.local_file.close()
        except FilesystemClosed:
            pass


class RegularBackupHandler(FilesystemHandler):
    """A customized FilesystemHandler that makes a copy to a remote location regularly after given amount of time.

    IMPORTANT: Make sure to combine this handler with a correct log filter. More in docstring of emit method.
    """

    def __init__(self, *args, backup_interval: Union[float, int], **kwargs):
        """
        :param backup_interval: A minimal number of seconds before handler will backup the log file to the remote
            location. The backup will only happen when the next log record will be emitted.
        """
        super().__init__(*args, **kwargs)

        self.backup_interval = backup_interval
        self._last_backup_time = time.monotonic()

    def emit(self, record: LogRecord) -> None:
        """Save a new record and backup to remote if the backup hasn't been done in the given amount of time.

        IMPORTANT: Any new log produced in this method must not reach this handler again. Otherwise the process will
        get stuck waiting for a thread lock release. Make sure that you combine this handler with a Filter class that
        filters out logs from botocore and s3transfer!
        """
        super().emit(record)

        if time.monotonic() > self._last_backup_time + self.backup_interval:
            self.local_file.copy_to_remote()
            self._last_backup_time = time.monotonic()


class EOExecutionHandler(FilesystemHandler):
    """A customized FilesystemHandler that makes a copy to a remote location every time a new node in a workflow
    is started.

    IMPORTANT: Make sure to combine this handler with a correct log filter. More in docstring of emit method.
    """

    def emit(self, record: LogRecord) -> None:
        """Save a new record. In case a new node in EOWorkflow is started it will copy the log file to remote.

        IMPORTANT: Any new log produced in this method must not reach this handler again. Otherwise the process will
        get stuck waiting for a thread lock release. Therefore, make sure that you combine this handler with a Filter
        class that filters out logs from botocore, s3transfer and anything else that could be possibly logged during
        `LocalFile.copy_to_remote` call!
        """
        super().emit(record)

        if record.name == "eolearn.core.eoworkflow" and record.message.startswith("Computing"):
            self.local_file.copy_to_remote()


class StdoutFilter(Filter):
    """Filters log messages passed to standard output"""

    DEFAULT_LOG_PACKAGES = (
        "eogrow",
        "__main__",
        "root",
        "sentinelhub.sentinelhub_batch",
    )

    def __init__(self, log_packages: Optional[Sequence[str]] = None, *args, **kwargs):
        """
        :param log_packages: Names of packages which logs to include.
        """
        super().__init__(*args, **kwargs)

        self.log_packages = self.DEFAULT_LOG_PACKAGES if log_packages is None else log_packages

    def filter(self, record: LogRecord) -> bool:
        """Shows only logs from eo-grow type packages and high-importance logs"""
        if record.levelno >= logging.WARNING:
            return True

        return any(package_name in record.name for package_name in self.log_packages)


class LogFileFilter(Filter):
    """Filters log messages passed to log file"""

    DEFAULT_IGNORE_PACKAGES = (
        "eolearn.core",
        "botocore",
        "s3transfer",
        "matplotlib",
        "fiona",
        "rasterio",
        "graphviz",
        "urllib3",
        "boto3",
    )

    def __init__(self, ignore_packages: Optional[Sequence[str]] = None, *args, **kwargs):
        """
        :param ignore_packages: Names of packages which logs will be ignored.
        """
        super().__init__(*args, **kwargs)

        self.ignore_packages = self.DEFAULT_IGNORE_PACKAGES if ignore_packages is None else ignore_packages

    def filter(self, record: LogRecord) -> bool:
        """Shows everything from the main thread and process except logs from packages that are on the ignore list.
        Those packages send a lot of useless logs.
        """
        if record.name.startswith(self.ignore_packages):
            return False

        return record.threadName == "MainThread" and record.processName == "MainProcess"


class EOExecutionFilter(Filter):
    """Filters logs that will be saved by EOExecutor"""

    DEFAULT_IGNORE_PACKAGES = (
        "botocore",
        "s3transfer",
        "urllib3",
        "rasterio",
        "numba",
        "fiona.ogrext",
    )

    def __init__(self, ignore_packages: Optional[Sequence[str]] = None, *args, **kwargs):
        """
        :param ignore_packages: Names of packages which logs will be ignored.
        """
        super().__init__(*args, **kwargs)

        self.ignore_packages = self.DEFAULT_IGNORE_PACKAGES if ignore_packages is None else ignore_packages

    def filter(self, record: LogRecord) -> bool:
        """Ignores logs from certain low-level packages"""
        if record.levelno >= logging.INFO:
            return True

        return not record.name.startswith(self.ignore_packages)



""" Includes custom types used in schemas
"""
from typing import Union, Tuple
import datetime

from eolearn.core import FeatureType

Path = str
S3Path = str
ImportPath = str
TimePeriod = Tuple[datetime.date, datetime.date]

Feature = Union[Tuple[FeatureType, str], FeatureType]
# Feature type is added for BBOX and TIMESTAMP, using Literal with Enum is weird



"""
Module defining common validators for schemas and validator wrappers
"""
import inspect
import datetime as dt
from typing import Callable, Tuple

from pydantic import validator

from sentinelhub import DataCollection

from ..utils.types import S3Path, TimePeriod


def field_validator(field: str, validator_fun: Callable, allow_reuse=True, **kwargs) -> Callable:
    """Sugared syntax for the `validator` decorator of `pydantic`"""
    return validator(field, allow_reuse=allow_reuse, **kwargs)(validator_fun)


def optional_field_validator(field: str, validator_fun: Callable, allow_reuse=True, **kwargs) -> Callable:
    """Wraps the validator functions so that `None` is always a valid input and only calls the validator on values.

    This allows re-use of validators e.g. if we have a validator for `Path` we can now use it for `Optional[Path]`.
    Because `pydantic` has a variable amount of arguments passed to the validator this function can only be used
    with validators that include `**kwargs` (or require all three arguments). For details on this behaviour
    consult the [validators documentation](https://pydantic-docs.helpmanual.io/usage/validators/).
    """
    # In order to propagate the pydantic python magic we need a bit of python magic ourselves
    additional_args = inspect.getfullargspec(validator_fun).args[1:]

    def optional_validator(value, values, config, field):
        if value is not None:
            all_kwargs = {"values": values, "config": config, "field": field}
            kwargs = {k: v for k, v in all_kwargs.items() if k in additional_args}
            return validator_fun(value, **kwargs)
        return None

    return validator(field, allow_reuse=allow_reuse, **kwargs)(optional_validator)


def validate_s3_path(value: S3Path) -> S3Path:
    """Validates the prefix of a S3 bucket path"""
    assert value.startswith("s3://"), "S3 path must start with s3://"
    return value


def parse_time_period(value: Tuple[str, str]) -> TimePeriod:
    """Allows parsing of preset options of shape `[preset_kind, year]` but that requires `pre` validation"""
    presets = ["yearly", "season", "Q1", "Q2", "Q3", "Q4", "Q1-yearly", "Q2-yearly", "Q3-yearly", "Q4-yearly"]

    if value[0] in presets:
        kind, year = value[0], int(value[1])
        start_dates = {
            "yearly": f"{year}-01-01",
            "season": f"{year-1}-09-01",
            "Q1": f"{year}-01-01",
            "Q2": f"{year}-04-01",
            "Q3": f"{year}-07-01",
            "Q4": f"{year}-10-01",
            "Q1-yearly": f"{year-1}-04-01",
            "Q2-yearly": f"{year-1}-07-01",
            "Q3-yearly": f"{year-1}-10-01",
            "Q4-yearly": f"{year}-01-01",
        }
        end_dates = {
            "yearly": f"{year}-12-31",
            "season": f"{year}-09-01",
            "Q1": f"{year}-03-31",
            "Q2": f"{year}-06-30",
            "Q3": f"{year}-09-30",
            "Q4": f"{year}-12-31",
            "Q1-yearly": f"{year}-03-31",
            "Q2-yearly": f"{year}-06-30",
            "Q3-yearly": f"{year}-09-30",
            "Q4-yearly": f"{year}-12-31",
        }
        value = start_dates[kind], end_dates[kind]

    start = dt.datetime.strptime(value[0], "%Y-%m-%d").date()
    end = dt.datetime.strptime(value[1], "%Y-%m-%d").date()
    assert start <= end, "Invalid start and end dates provided. End date must follow the start date"
    return start, end


def parse_data_collection(value: str) -> DataCollection:
    """Validates and parses data collection"""
    if value.startswith("BYOC_"):
        collection_id = value.split("_")[-1]
        return DataCollection.define_byoc(collection_id)

    if value.startswith("BATCH_"):
        collection_id = value.split("_")[-1]
        return DataCollection.define_batch(collection_id)

    if value in DataCollection.__members__:
        return getattr(DataCollection, value)
    raise ValueError(
        "Data collection should be a name of an existing DataCollection enum, 'BYOC_<collection_id>', "
        f"or 'BATCH_<collection id>' but name '{value}' was given."
    )


def validate_resampling(value: str) -> str:
    """Ensure that the resampling parameter has a valid value"""
    assert value in ["NEAREST", "BILINEAR", "BICUBIC"]
    return value


def validate_mosaicking_order(value: str) -> str:
    """Ensure that the mosaicking_order parameter has a valid value"""
    assert value in ["mostRecent", "leastRecent", "leastCC"]
    return value



"""
Pipelines that transform data
"""
from typing import Dict
from pydantic import Field
from eolearn.core import LoadTask, EOWorkflow, linearly_connect_tasks, SaveTask, FeatureType, OverwritePermission

from ..core.pipeline import Pipeline
from ..tasks.common import MappingTask


class MappingPipeline(Pipeline):
    """A pipeline for transforming a values of a mask_timeless feature according to given mapping rules"""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(description="The storage manager key pointing to the input folder")
        input_feature: str = Field(description="Name of the input mask_timeless feature")
        output_folder_key: str = Field(description="The storage manager key pointing to the output folder")
        output_feature: str = Field(description="Name of the output mask_timeless feature")

        mapping_dictionary: Dict[int, int] = Field(description="Mapping dictionary of the input-to-output classes")
        compress_level: int = Field(1, description="Level of compression used in saving eopatches")

    def build_workflow(self) -> EOWorkflow:
        """Method for constructing the workflow"""
        input_feature = FeatureType.MASK_TIMELESS, self.config.input_feature
        output_feature = FeatureType.MASK_TIMELESS, self.config.output_feature
        extra_features = [FeatureType.BBOX] if self.config.input_folder_key != self.config.output_folder_key else []

        load_task = LoadTask(
            self.storage.get_folder(self.config.input_folder_key, full_path=True),
            features=[input_feature] + extra_features,
            config=self.sh_config,
        )

        mapping_task = MappingTask(input_feature, output_feature, self.config.mapping_dictionary)

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key, full_path=True),
            features=[output_feature] + extra_features,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            config=self.sh_config,
            compress_level=self.config.compress_level,
        )

        return EOWorkflow(linearly_connect_tasks(load_task, mapping_task, save_task))



"""
Module containing utilities for working with filesystems
"""
from typing import Optional

import fs
import fs.copy
from fs.osfs import OSFS
from fs.tempfs import TempFS

from sentinelhub import SHConfig
from eolearn.core.utils.fs import get_base_filesystem_and_path


class LocalFile:
    """An abstraction for working with a local version of a file.

    If file's original location is remote (e.g. on S3) then this will ensure working with a local copy. If file's
    original location is already on local filesystem then it will work with that, unless `always_copy=True` is used.
    """

    def __init__(
        self,
        path: str,
        mode: str = "r",
        filesystem: Optional[fs.base.FS] = None,
        config: Optional[SHConfig] = None,
        always_copy: bool = False,
        **temp_fs_kwargs,
    ):
        """
        :param path: Either a full path to a remote file or a path to remote file which is relative to given filesystem
            object.
        :type path: str
        :param mode: One of the option `r', 'w', and 'rw`, which specify if a file should be read or written to remote.
            The default is 'r'.
        :type mode: str
        :param filesystem: A filesystem of the remote. If not given, it will be determined from the path.
        :type: fs.FS
        :param config: A config object with which AWS credentials could be used to initialize a remote filesystem object
        :type config: SHConfig
        :param always_copy: If True it will always make a local copy to a temporary folder, even if a file is already
            in the local filesystem.
        :type always_copy: bool
        :param temp_fs_kwargs: Parameters that will be propagated to fs.tempfs.TempFS
        """
        if filesystem is None:
            filesystem, path = get_base_filesystem_and_path(path, config=config)
        self._remote_path = path
        self._remote_filesystem = filesystem

        if not (mode and isinstance(mode, str) and set(mode) <= {"r", "w"}):
            raise ValueError(f"Parameter mode should be one of the strings 'r', 'w' or 'rw' but {mode} found")
        self._mode = mode

        if isinstance(self._remote_filesystem, (OSFS, TempFS)) and not always_copy:
            self._filesystem = self._remote_filesystem
            self._local_path = self._remote_path
        else:
            self._filesystem = TempFS(**temp_fs_kwargs)
            self._local_path = fs.path.basename(self._remote_path)

        self._absolute_local_path = self._filesystem.getsyspath(self._local_path)

        if "r" in self._mode:
            self.copy_to_local()
        elif "w" in self._mode:
            remote_dirs = fs.path.dirname(self._remote_path)
            self._remote_filesystem.makedirs(remote_dirs, recreate=True)

    @property
    def path(self) -> str:
        """Provides an absolute path to the copy in the local filesystem"""
        return self._absolute_local_path

    def __enter__(self):
        """This allows the class to be used as a context manager"""
        return self

    def __exit__(self, *_, **__):
        """This allows the class to be used as a context manager. In case an error is raised this will by default
        still delete the object from local folder. In case you don't want that, initialize `LocalFile` with
        `auto_clean=False`.
        """
        self.close()

    def close(self):
        """Close the local copy"""
        if "w" in self._mode:
            self.copy_to_remote()

        if self._filesystem is not self._remote_filesystem:
            self._filesystem.close()

    def copy_to_local(self):
        """Copy from remote to local location"""
        if self._filesystem is not self._remote_filesystem:
            fs.copy.copy_file(self._remote_filesystem, self._remote_path, self._filesystem, self._local_path)

    def copy_to_remote(self):
        """Copy from local to remote location"""
        if self._filesystem is not self._remote_filesystem:
            fs.copy.copy_file(self._filesystem, self._local_path, self._remote_filesystem, self._remote_path)



"""
Module implementing merge samples
"""
import logging
from typing import List, Optional, Sequence, Tuple, Literal

import fs
import numpy as np
from pydantic import Field

from eolearn.core import FeatureType, linearly_connect_tasks, LoadTask, OutputTask, EOWorkflow, EOPatch
from eolearn.core.utils.fs import get_full_path
from eolearn.core.utils.parsing import parse_features

from ..core.pipeline import Pipeline
from ..utils.types import Feature

LOGGER = logging.getLogger(__name__)


class MergeSamplesPipeline(Pipeline):
    """Pipeline to merge sampled data into joined numpy arrays"""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(
            description="The storage manager key pointing to the input folder for the merge samples."
        )
        output_folder_key: str = Field(
            description="The storage manager key pointing to the output folder for the merge samples pipeline."
        )
        features_to_merge: List[Feature] = Field(
            default_factory=dict, description="Dictionary of all features for which samples are to be merged."
        )
        id_filename: Optional[str] = Field(description="Filename of array holding patch id of concatenated features")
        suffix: str = Field("", description="String to append to array filenames")
        workers: int = Field(1, description="Number of threads used to load data from EOPatches in parallel.")
        use_ray: Literal[False] = Field(False, description="Pipeline does not parallelize properly.")

    _OUTPUT_NAME = "features-to-merge"

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Procedure which merges data from EOPatches into ML-ready numpy arrays"""
        features_to_merge = self._get_features_to_merge()

        workflow = self.build_workflow(features_to_merge)
        exec_args = self.get_execution_arguments(workflow)

        # It doesn't make sense to parallelize loading over a cluster, but it would # make sense to parallelize over
        # features that have to be concatenated or, if we would concatenate into multiple files, parallelize creating
        # batches of features
        successful, failed, results = self.run_execution(workflow, exec_args, multiprocess=False)

        result_patches = [result.outputs.get(self._OUTPUT_NAME) for result in results]

        self.merge_and_save_features(result_patches, features_to_merge, patch_names=successful)

        return successful, failed

    def _get_features_to_merge(self) -> List[Feature]:
        """Creates a list of feature to be merged"""
        merge_features = parse_features(
            self.config.features_to_merge,
            allowed_feature_types=[
                FeatureType.DATA,
                FeatureType.DATA_TIMELESS,
                FeatureType.MASK,
                FeatureType.MASK_TIMELESS,
                FeatureType.TIMESTAMP,
            ],
        )

        if not merge_features:
            raise ValueError("At least one feature should be given to merge")

        if any(FeatureType.TIMESTAMP in feature for feature in merge_features):
            merge_features = [feature for feature in merge_features if FeatureType.TIMESTAMP not in feature]
            # It is important that this comes after other features to tile it correctly
            merge_features.append((FeatureType.TIMESTAMP, None))

        return merge_features

    def build_workflow(self, features: Sequence[Feature]) -> EOWorkflow:
        """Creates a workflow that outputs the requested features"""
        load_task = LoadTask(
            self.storage.get_folder(self.config.input_folder_key, full_path=True),
            features=features,
            config=self.sh_config,
        )
        output_task = OutputTask(name=self._OUTPUT_NAME)
        return EOWorkflow(linearly_connect_tasks(load_task, output_task))

    def merge_and_save_features(
        self, patches: List[EOPatch], features_to_merge: List[Feature], patch_names: List[str]
    ) -> None:
        """Merges features from EOPatches and saves data"""
        patch_sample_nums = None
        for feature in features_to_merge:
            LOGGER.info("Started merging feature %s", feature)
            arrays = [self._collect_and_remove_feature(patch, feature) for patch in patches]
            feature_type, feature_name = feature

            if patch_sample_nums is None:
                patch_sample_nums = [array.shape[0] for array in arrays]

            if feature_type is FeatureType.TIMESTAMP:
                arrays = [
                    np.tile(timestamp, (sample_num, 1)) for sample_num, timestamp in zip(patch_sample_nums, arrays)
                ]
                feature_name = "TIMESTAMPS"

            merged_array = np.concatenate(arrays, axis=0)
            del arrays

            self._save_array(merged_array, feature_name)
            del merged_array

        if self.config.id_filename:
            LOGGER.info("Started merging EOPatch ids")
            patch_ids = self.eopatch_manager.get_id_list_from_eopatch_list(patch_names)
            patch_id_arrays = [
                np.ones(sample_num, dtype=np.uint32) * patch_id
                for sample_num, patch_id in zip(patch_sample_nums, patch_ids)
            ]

            merged_patch_ids = np.concatenate(patch_id_arrays, axis=0)
            self._save_array(merged_patch_ids, self.config.id_filename)

    @staticmethod
    def _collect_and_remove_feature(patch: EOPatch, feature: Feature) -> np.ndarray:
        """Collects a feature from an EOPatch and removes it from EOPatch to conserve overall memory"""
        feature_array = patch[feature]
        feature_type, _ = feature

        if feature_type is FeatureType.TIMESTAMP:
            patch.timestamp = []
            return np.array(feature_array)

        del patch[feature]

        axis = feature_type.ndim() - 2
        feature_array = np.squeeze(feature_array, axis=axis)

        if feature_type in [FeatureType.DATA, FeatureType.MASK]:
            feature_array = np.moveaxis(feature_array, 0, 1)

        if feature_array.dtype == np.float64:
            feature_array = feature_array.astype(np.float32)

        return feature_array

    def _save_array(self, array: np.ndarray, name: str) -> None:
        """Saves an array to the storage with training data using the name of the array"""
        save_folder = self.storage.get_folder(self.config.output_folder_key)
        if self.config.suffix:
            name = f"{name}_{self.config.suffix}"
        path = fs.path.combine(save_folder, f"{name}.npy")

        with self.storage.filesystem.openbin(path, "w") as file_handle:
            np.save(file_handle, array)

        LOGGER.info("Saved concatenated array to %s", get_full_path(self.storage.filesystem, path))



"""
Utilities for filtering eopatch lists
"""
from typing import List, Iterable
from concurrent.futures import ThreadPoolExecutor

import fs
from fs.base import FS

from eolearn.core.eodata_io import walk_filesystem

from ..utils.types import Feature


def check_if_features_exist(
    filesystem: FS,
    eopatch_path: str,
    features: Iterable[Feature],
) -> bool:
    """Checks whether an EOPatch in the given location has all specified features saved"""
    not_seen_features = set(features)
    try:
        for (ftype, name, _) in walk_filesystem(filesystem, eopatch_path, features=features):
            if (ftype, name) in not_seen_features:
                not_seen_features.remove((ftype, name))
            elif ftype in not_seen_features:
                not_seen_features.remove(ftype)
    except (IOError, fs.errors.ResourceNotFound):
        return False
    return len(not_seen_features) == 0


def get_patches_without_all_features(
    filesystem: FS,
    patches_folder: str,
    patch_list: List[str],
    features: Iterable[Feature],
) -> List[str]:
    """Filters out patches that already contain all features"""
    eopatch_paths = [fs.path.combine(patches_folder, eopatch) for eopatch in patch_list]

    def check_patch(eopatch_path):
        return check_if_features_exist(filesystem, eopatch_path, features)

    with ThreadPoolExecutor() as pool:
        has_features_iter = pool.map(check_patch, eopatch_paths)

    return [eopatch for eopatch, has_features in zip(patch_list, has_features_iter) if not has_features]



"""
Utilities for parallelization of the code
"""
from typing import Optional, Callable
import concurrent.futures


def parallelize_with_threads(function: Callable, *params, workers: Optional[int] = None) -> list:
    """A simple parallelization helper function"""
    with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
        return list(executor.map(function, *params))



"""
A pipeline module for rasterizing vector datasets.
"""
import os
import logging
import uuid
from typing import List, Optional, Union, Tuple

import fiona
import fs
import numpy as np
import geopandas as gpd
from pydantic import Field, validator
from eolearn.core import (
    EOWorkflow,
    FeatureType,
    CreateEOPatchTask,
    LoadTask,
    SaveTask,
    OverwritePermission,
    EONode,
    MergeEOPatchesTask,
)
from eolearn.io import VectorImportTask
from eolearn.geometry import VectorToRasterTask
from eolearn.core.utils.fs import join_path
from eolearn.core.utils.parsing import parse_feature

from ..core.config import Config
from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..utils.fs import LocalFile
from ..utils.vector import concat_gdf
from ..utils.types import Feature
from ..utils.filter import get_patches_without_all_features

LOGGER = logging.getLogger(__name__)


class VectorColumnSchema(BaseSchema):
    """Parameter structure for individual feature / dataset column to be rasterized"""

    value: Optional[float] = Field(
        description="Value to use for all rasterized polygons. Use either this or `values_column`."
    )
    values_column: Optional[str] = Field(
        description=(
            "GeoPandas dataframe column name from which to read values for geometries. Use either this or `value`."
        )
    )
    output_feature: Feature = Field(description="Output feature of rasterization.")
    polygon_buffer: float = Field(0, description="The size of polygon buffering to be applied before rasterization.")
    resolution: float = Field(description="Rendering resolution in meters.")
    overlap_value: Optional[int] = Field(description="Value to write over the areas where polygons overlap.")
    dtype: str = Field("int32", description="Numpy dtype of the output feature.")
    no_data_value: int = Field(0, description="The no_data_value argument to be passed to VectorToRasterTask")

    @validator("values_column")
    def check_value_settings(cls, v, values):  # pylint ignore:invalid-name,no-self-use,no-self-argument
        """Ensures that precisely one of `value` and `values_column` is set."""
        assert (v is None) != (values["value"] is None), "Only one of `values_column` and `value` should be given."
        return v


class Preprocessing(BaseSchema):
    reproject_crs: Optional[int] = Field(
        description=(
            "An EPSG code of a CRS in which vector_input data will be reprojected once loaded. This parameter "
            "is mandatory if vector_input contains multiple layers in different CRS"
        ),
    )


class RasterizePipeline(Pipeline):
    """A pipeline module for rasterizing vector datasets."""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(
            description="The storage manager key pointing to the input folder for the rasterization pipeline."
        )
        output_folder_key: str = Field(
            description="The storage manager key pointing to the output folder for the rasterization pipeline pipeline."
        )
        vector_input: Union[Feature, str] = Field(
            description="A filename in the input_data folder or a feature containing vector data."
        )
        dataset_layer: Optional[str] = Field(
            description="Name of a layer with data to be rasterized in a multi-layer file."
        )
        columns: List[VectorColumnSchema]
        preprocess_dataset: Optional[Preprocessing] = Field(
            description=(
                "Parameters used by self.preprocess_dataset method. If set to `None` it skips the dataframe preprocess"
                " step."
            )
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self._is_input_file(self.config.vector_input):
            self.vector_feature = FeatureType.VECTOR_TIMELESS, f"TEMP_{uuid.uuid4().hex}"
        else:
            self.vector_feature = parse_feature(self.config.vector_input)

    def filter_patch_list(self, patch_list: List[str]) -> List[str]:
        filtered_patch_list = get_patches_without_all_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            self._get_output_features(),
        )

        return filtered_patch_list

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        if self._is_input_file(self.config.vector_input) and self.config.preprocess_dataset is not None:
            self.run_dataset_preprocessing()
        return super().run_procedure()

    def run_dataset_preprocessing(self) -> None:
        """Loads datasets, applies preprocessing steps and saves them to a cache folder"""
        LOGGER.info("Preprocessing dataset %s", self.config.vector_input)

        file_path = fs.path.combine(self.storage.get_input_data_folder(), self.config.vector_input)
        with LocalFile(file_path, mode="r", filesystem=self.storage.filesystem) as local_file:
            dataset_layers = [
                gpd.read_file(local_file.path, layer=layer, encoding="utf-8")
                for layer in fiona.listlayers(local_file.path)
            ]

        dataset_gdf = concat_gdf(dataset_layers, reproject_crs=self.config.preprocess_dataset.reproject_crs)

        dataset_gdf = self.preprocess_dataset(dataset_gdf)

        dataset_path = self._get_dataset_path(self.config, full_path=False)
        with LocalFile(dataset_path, mode="w", filesystem=self.storage.filesystem) as local_file:
            dataset_gdf.to_file(local_file.path, encoding="utf-8", driver="GPKG")

    def build_workflow(self) -> EOWorkflow:
        """Creates workflow that is divided into the following sub-parts:

        1. loading data,
        2. preprocessing steps,
        3. rasterization of features,
        4. postprocessing steps,
        5. saving results
        """
        if self._is_input_file(self.config.vector_input):
            create_node = EONode(CreateEOPatchTask())
            path = self._get_dataset_path(self.config)
            import_task = VectorImportTask(
                self.vector_feature, path=path, layer=self.config.dataset_layer, config=self.sh_config
            )
            data_preparation_node = EONode(import_task, inputs=[create_node])
        else:
            input_task = LoadTask(
                self.storage.get_folder(self.config.input_folder_key, full_path=True),
                features=[self.vector_feature, FeatureType.BBOX],
                config=self.sh_config,
            )
            data_preparation_node = EONode(input_task)

        preprocess_node = self.get_prerasterization_node(data_preparation_node)

        rasterization_node = self.get_rasterization_node(preprocess_node)

        postprocess_node = self.get_postrasterization_node(rasterization_node)

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key, full_path=True),
            features=self._get_output_features(),
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            config=self.sh_config,
        )
        save_node = EONode(save_task, inputs=[postprocess_node])

        return EOWorkflow.from_endnodes(save_node)

    def preprocess_dataset(self, dataframe: gpd.GeoDataFrame) -> gpd.GeoDataFrame:  # pylint: disable=no-self-use
        """Method for applying custom preprocessing steps on the entire dataset"""
        return dataframe

    def get_prerasterization_node(self, previous_node: EONode) -> EONode:  # pylint: disable=no-self-use
        """Builds node with tasks to be applied after loading vector feature but before rasterization"""
        return previous_node

    def get_rasterization_node(self, previous_node: EONode) -> EONode:
        """Builds nodes containing rasterization tasks"""

        rasterization_nodes = [
            EONode(
                inputs=[previous_node],
                task=VectorToRasterTask(
                    vector_input=self.vector_feature,
                    raster_feature=column.output_feature,
                    values_column=column.values_column,
                    values=column.value,
                    buffer=column.polygon_buffer,
                    raster_resolution=column.resolution,
                    raster_dtype=np.dtype(column.dtype),
                    no_data_value=column.no_data_value,
                    overlap_value=column.overlap_value,
                ),
            )
            for column in self.config.columns
        ]

        if len(rasterization_nodes) == 1:
            return rasterization_nodes[0]
        return EONode(MergeEOPatchesTask(), inputs=rasterization_nodes)

    def get_postrasterization_node(self, previous_node: EONode) -> EONode:  # pylint: disable=no-self-use
        """Builds node with tasks to be applied after rasterization"""
        return previous_node

    @staticmethod
    def _is_input_file(value: Union[Feature, str]) -> bool:
        """Checks if given name ends with one of the supported file extensions"""
        return isinstance(value, str) and value.lower().endswith((".geojson", ".shp", ".gpkg", ".gdb"))

    def _get_dataset_path(self, config: Config, full_path: bool = True) -> str:
        """Provides a path from where dataset should be loaded into the workflow"""
        if config.preprocess_dataset is not None:
            folder = self.storage.get_cache_folder(full_path=full_path)
            filename = f"preprocessed_{config.vector_input}"
            filename = (os.path.splitext(filename))[0] + ".gpkg"
        else:
            folder = self.storage.get_input_data_folder(full_path=full_path)
            filename = config.vector_input

        if full_path:
            return join_path(folder, filename)
        return fs.path.combine(folder, filename)

    def _get_output_features(self) -> List[Feature]:
        """Lists all features that are to be saved upon the pipeline completion"""
        base_features = [FeatureType.BBOX]
        rasterized_features = [column.output_feature for column in self.config.columns]
        return base_features + rasterized_features



"""
Module for handling EOPatch naming conventions
"""
import json
import os
from typing import Optional, List, Union, cast, Dict
import warnings

import bidict
import pandas
from geopandas import GeoDataFrame
from pydantic import Field

from sentinelhub import BBox, CRS

from .area.base import AreaManager
from .base import EOGrowObject
from .config import Config
from .schemas import ManagerSchema


class EOPatchManager(EOGrowObject):
    """Class defining and handling naming convention of EOPatches that divide AOI"""

    class Schema(ManagerSchema):
        pass

    def __init__(self, config: Config, area_manager: AreaManager):
        """
        :param config: Configuration of the manager
        :param area_manager: Area managing class containing info about how the area is split into EOPatches
        """
        super().__init__(config)

        self._area_manager = area_manager

        self._name_to_id_map = None
        self._name_to_bbox_map = None

    @property
    def name_to_id_map(self) -> bidict.bidict:
        """Provides a bidirectional mapping between EOPatch names and EOPatch ids. The mapping is generated lazily the
        first time it is needed.
        """
        if self._name_to_id_map is None:
            self._prepare_names()
        return self._name_to_id_map  # type: ignore

    @property
    def name_to_bbox_map(self) -> dict:
        """Provides a dictionary mapping EOPatch names to bounding boxes."""
        if self._name_to_bbox_map is None:
            self._prepare_names()
        return self._name_to_bbox_map  # type: ignore

    def _prepare_names(self):
        """Collects bounding boxes from an instance of AreaManager class, generates EOPatch names and saves them into
        a private class variable _eopatch_name_list
        """
        bbox_grid = self._area_manager.get_grid(add_bbox_column=True)
        if not bbox_grid:
            return

        bbox_df = pandas.concat(bbox_grid, ignore_index=True)
        bbox_df = bbox_df.sort_values(by="index_n")

        if self._area_manager.has_region_filter():
            total_patch_num = self._area_manager.get_grid_size(ignore_region_filter=True)
        else:
            total_patch_num = len(bbox_df.index)

        self._name_to_id_map = self.generate_names(bbox_df, total_patch_num=total_patch_num)
        self._name_to_bbox_map = dict(zip(self._name_to_id_map, bbox_df["BBOX"]))

    @staticmethod
    def generate_names(bbox_dataframe: GeoDataFrame, total_patch_num: Optional[int] = None) -> bidict.bidict:
        """Method that generates EOPatch names from a dataframe holding bounding boxes. This method can be overridden
        by a method that generates names in a different way.

        :param bbox_dataframe: A dataframe with bounding box geometries and information
        :param total_patch_num: Total number of EOPatches
        :return: A bidirectional dictionary between names and original IDs
        """
        if total_patch_num is None:
            total_patch_num = len(bbox_dataframe.index)
        zfill_length = len(str(total_patch_num - 1))

        return bidict.bidict(
            (f"eopatch-id-{row.index_n:0{zfill_length}}-col-{row.index_x}-row-{row.index_y}", int(row.index_n))
            for _, row in bbox_dataframe.iterrows()
        )

    def is_eopatch_name(self, name: str) -> bool:
        """Checks if the given name (could be entire file path) is the name of an EOPatch

        :param name: A name or a file path of a folder which could be one of EOPatches
        :type name: str
        """
        return os.path.basename(name) in self.name_to_id_map

    def get_eopatch_filenames(
        self, folder: Optional[str] = None, id_list: Optional[List[int]] = None, filter_existing: bool = True
    ) -> List[str]:
        """Provides a list of EOPatch names

        :param folder: If it is specified it will join the folder with eopatch names and return a list of complete paths
        :param id_list: A list of patch ids which should be provided. By default is set to None and all EOPatch names
            will be provided
        :param filter_existing: If given a folder parameter this will check if EOPatches at the given location actually
            exist. By default this is set to True which means it will make additional IO calls, which could be slow
            if using s3.
        :return: A list of EOPatch folder names
        """
        filenames = self._eopatch_list_from_id_list(id_list)  # TODO

        if folder is None:
            return filenames

        file_paths = [os.path.join(folder, name) for name in filenames]

        if filter_existing:
            file_paths = [path for path in file_paths if os.path.isdir(path)]

            total_count, filtered_count = len(filenames), len(file_paths)
            if filtered_count < total_count:
                warnings.warn(
                    f"Only {filtered_count} out of {total_count} EOPatches found in folder {folder}", RuntimeWarning
                )
        return file_paths

    def save_eopatch_filenames(self, filename: str, eopatch_list: Union[List[int], List[str], None] = None):
        """Saves a list of EOPatches to a file

        :param filename: A filename (or entire file path) where names of EOPatches will be saved. Supported formats
            are JSON and TXT
        :param eopatch_list: Can either be a list of EOPatch names (or file paths) or a list of indices. If this is not
            specified, all names will be saved
        """
        if eopatch_list is None:
            eopatch_list = list(self.name_to_id_map)
        eopatch_list = self.parse_eopatch_list(eopatch_list)

        if filename.endswith(".json"):
            with open(filename, "w") as file:
                json.dump(eopatch_list, file, indent=2)
        elif filename.endswith(".txt"):
            with open(filename, "w") as file:
                print("\n".join(eopatch_list), end="", file=file)
        else:
            raise ValueError(f"Unrecognized file format of {filename}")

    def load_eopatch_filenames(self, filename: str, id_list: Optional[List[int]] = None) -> List[str]:
        """Loads a list of EOPatch names from a file

        :param filename: A filename (or entire file path) from where names of EOPatches will be loaded. Supported
            formats are JSON and TXT
        :param id_list: A list of EOPatch IDs which are the only ones required. The IDs are calculated according to
            the complete list of EOPatches. By default no filtering is done.
        :return: A list of EOPatch names loaded from file (and maybe filtered)
        """
        if filename.endswith(".json"):
            with open(filename, "r") as file:
                eopatch_list = json.load(file)
        elif filename.endswith(".txt"):
            with open(filename, "r") as file:
                content = file.read()
                eopatch_list = [name.strip(" \t") for name in content.replace("\n", ",").split(",")]
                eopatch_list = [(int(name) if name.isdigit() else name) for name in eopatch_list if name]
        else:
            raise ValueError(f"Unrecognized file format of {filename}")

        eopatch_list = self.parse_eopatch_list(eopatch_list)

        if id_list is not None:
            filter_set = set(self._eopatch_list_from_id_list(id_list))
            eopatch_list = [name for name in eopatch_list if name in filter_set]

        return eopatch_list

    def get_bboxes(self, eopatch_list: Optional[List[str]] = None) -> List[BBox]:  # TODO: add unit test
        """Provides bounding boxes for each EOPatch

        :param eopatch_list:
        :return: A list of bounding box objects
        """
        if eopatch_list is None:
            return list(self.name_to_bbox_map.values())

        eopatch_list = self.parse_eopatch_list(eopatch_list)
        return [self.name_to_bbox_map[eopatch_name] for eopatch_name in eopatch_list]

    def parse_eopatch_list(self, eopatch_list: Union[List[str], List[int]]) -> List[str]:
        """Parses given list of EOPatches into a a standard format

        :param eopatch_list: Can either be a list of EOPatch names (or file paths) or a list of indices
        """
        if not isinstance(eopatch_list, list):
            raise ValueError(f"Expected a list of EOPatch names, got {eopatch_list}")

        if all(isinstance(name, str) for name in eopatch_list):
            eopatch_list = cast(List[str], eopatch_list)  # informs mypy of the type, doesnt change anything
            eopatch_list = [os.path.basename(name) for name in eopatch_list]

            if set(eopatch_list) <= self.name_to_id_map.keys():
                return eopatch_list

            raise ValueError("The list of EOPatch names contains invalid names")

        if all(isinstance(name, int) for name in eopatch_list):
            return [self.name_to_id_map.inverse[idx] for idx in eopatch_list]

        raise ValueError(
            f"Elements of a list of EOPatch names should be all strings or all integers, got {eopatch_list}"
        )

    def _eopatch_list_from_id_list(self, id_list: Optional[List[int]]) -> List[str]:
        """A method that provides EOPatch list from id list"""
        if id_list is None:
            return list(self.name_to_id_map)

        return [self.name_to_id_map.inverse[eopatch_id] for eopatch_id in id_list]

    def get_id_list_from_eopatch_list(self, eopatch_list: Optional[List[str]] = None) -> List[int]:
        """Return patch ID given the eopatch name"""
        if eopatch_list is None:
            return list(self.name_to_id_map.values())
        return [self.name_to_id_map[eopatch] for eopatch in eopatch_list]

    def split_by_utm(self, eopatch_list: list) -> Dict[CRS, list]:
        """Creates dict of eopatches by utm."""
        crs_eopatch_dict: Dict[CRS, list] = {}
        for eopatch in eopatch_list:
            crs = self.name_to_bbox_map[eopatch].crs
            crs_eopatch_dict[crs] = crs_eopatch_dict.get(crs, [])
            crs_eopatch_dict[crs].append(eopatch)

        return crs_eopatch_dict


class CustomGridEOPatchManager(EOPatchManager):
    """A custom patch manager that uses a naming convention defined in a grid"""

    class Schema(EOPatchManager.Schema):
        name_column: str = Field(description="A name of a column in grid dataframes that contains EOPatch names")
        index_column: str = Field(description="A name of a column in grid dataframes that contains EOPatch indices")

    def generate_names(self, bbox_dataframe: GeoDataFrame, *args, **kwargs) -> bidict.bidict:
        """Creates a bidirectional dictionary between names and indices"""
        names = bbox_dataframe[self.config.name_column]
        indices = bbox_dataframe[self.config.index_column]
        return bidict.bidict((name, index) for name, index in zip(names, indices))


class BatchTileManager(EOPatchManager):
    """A custom patch manager that uses a naming convention based on Sentinel Hub Batch tiles"""

    @staticmethod
    def generate_names(bbox_dataframe: GeoDataFrame, *args, **kwargs):
        """Creates a bidirectional dictionary between names and indices"""
        return bidict.bidict((name, index) for name, index in zip(bbox_dataframe.name, bbox_dataframe.index_n))



"""
Common tasks shared between pipelines
"""
from typing import List, Optional, Union, Callable

import numpy as np
from eolearn.core import EOPatch, EOTask, MapFeatureTask
from eolearn.geometry import MorphologicalOperations

from ..utils.types import Feature


class MappingTask(MapFeatureTask):
    """A task that takes an input mask_timeless feature and creates an output mask_timeless feature, where the
    input mask values are mapped to the output mask values, using the provided mapping dictionary.
    """

    def __init__(self, input_feature: Feature, output_feature: Feature, mapping_dict: dict):
        super().__init__(input_feature, output_feature, mapping_dict=mapping_dict)

    def map_method(self, feature: np.ndarray, mapping_dict: dict) -> np.ndarray:
        mapped_values = feature.copy()
        for map_from, map_to in mapping_dict.items():
            mapped_values[feature == map_from] = map_to

        return mapped_values


class ClassFilterTask(EOTask):
    """
    Run class specific morphological operation.
    """

    def __init__(
        self,
        feature: Feature,
        labels: List[int],
        morph_operation: Union[MorphologicalOperations, Callable],
        struct_elem: Optional[np.ndarray] = None,
    ):
        """Perform a morphological operation on a given feature mask

        :param feature: Feature to be modified
        :param labels: List of labels to be considered for morphological operation
        :param morph_operation: Type of morphological operation ot perform
        :param struct_elem: Structured element to be used. Taken from `ml_tools.MorphologicalStructFactory`
        """
        self.feature_type, self.feature_name, self.new_feature_name = self.parse_renamed_feature(feature)
        self.labels = labels

        if isinstance(morph_operation, MorphologicalOperations):
            self.morph_operation = MorphologicalOperations.get_operation(morph_operation)
        else:
            self.morph_operation = morph_operation
        self.struct_elem = struct_elem

    def execute(self, eopatch: EOPatch) -> EOPatch:
        mask = eopatch[self.feature_type][self.feature_name].copy()

        for label in self.labels:
            label_mask = np.squeeze((mask == label), axis=-1)
            mask_mod = self.morph_operation(label_mask, self.struct_elem) * label
            mask_mod = mask_mod[..., np.newaxis]
            mask[mask == label] = mask_mod[mask == label]

        eopatch[self.feature_type][self.new_feature_name] = mask
        return eopatch



"""
Module defining a base pipeline schema and custom fields

For each pipeline a separate schema has to be defined which inherits from PipelineSchema. Such schema should be placed
as an internal class of the implemented pipeline class
"""
from typing import Optional, List, Type, Dict, Union, Literal
from inspect import isclass

from pydantic import BaseModel, Field
from pydantic.fields import ModelField

from ..utils.types import Path, ImportPath


class BaseSchema(BaseModel):
    """A BaseModel that does not allow extra fields and checks default values."""

    class Config:
        extra = "forbid"
        validate_all = True


class RawManagerSchema(BaseModel):
    """A schema for managers that delay manager config verification, should not be used outside of `PipelineSchema`"""

    manager: ImportPath = Field(description="Import path to an implementation of a manager class.")

    class Config:
        extra = "allow"


class ManagerSchema(BaseSchema):
    """A basic schema for managers, to be used as a parent class for defining manager schemas"""

    manager: Optional[ImportPath] = Field(description="An import path to this specific manager.")


class PipelineSchema(BaseSchema):
    """Pipeline schema"""

    pipeline: Optional[ImportPath] = Field(description="Import path to an implementation of Pipeline class.")
    storage: RawManagerSchema = Field(description="A schema of an implementation of StorageManager class")
    area: RawManagerSchema = Field(description="A schema of an implementation of AreaManager class")
    eopatch: RawManagerSchema = Field(description="A schema of an implementation of EOPatchManager class")
    logging: RawManagerSchema = Field(description="A schema of an implementation of LoggingManager class")

    workers: int = Field(
        1, description="Number of workers for parallel execution of workflows. Parameter does not affect ray clusters."
    )
    use_ray: Union[Literal["auto"], bool] = Field(
        "auto",
        description=(
            "Whether to run the pipeline locally or using a (local or remote) ray cluster. When using `auto` the"
            " pipeline checks if it can connect to a cluster, and if none are available runs locally."
        ),
    )

    patch_list: Optional[List[int]] = Field(
        description="A list of EOPatch indices for which the pipeline should be executed"
    )
    input_patch_file: Optional[Path] = Field(description="File path to a file with input EOPatch names")
    skip_existing: bool = Field(
        False,
        description=(
            "Whether or not to skip already processed patches. In order to use this functionality a pipeline "
            "must implement the `filter_patch_list` method."
        ),
    )


def build_schema_template(schema: Type[BaseModel]) -> dict:
    """From a given schema class it creates a template of a config file. It does that by modifying
    OpenAPI-style schema.
    """
    openapi_schema = schema.schema()

    template_mapping: dict = {}
    model_schemas = openapi_schema.get("definitions", {})
    for name, model_schema in model_schemas.items():
        model_template = _process_model_schema(model_schema, template_mapping)
        template_mapping[name] = model_template

    return _process_model_schema(openapi_schema, template_mapping)


def _process_model_schema(openapi_schema: dict, template_mapping: Dict[str, dict]) -> dict:
    """Processes schema for a single model into a template"""
    template = _get_basic_template(openapi_schema)

    params = openapi_schema.get("properties", {})
    required_params = set(openapi_schema.get("required", []))

    for param_name, param_schema in params.items():
        param_template = _get_basic_template(param_schema)

        if param_name in required_params:
            param_template["#required"] = True

        referred_model_name = _get_referred_model_name(param_schema)
        if referred_model_name:
            # In case param_template and referred_model have some parameters in common the following prioritizes the
            # ones from param_template
            for model_param_name, model_param_value in template_mapping[referred_model_name].items():
                if model_param_name not in param_template:
                    param_template[model_param_name] = model_param_value

        template[param_name] = param_template

    return template


_SUPPORTED_OPENAPI_FIELDS = {"title", "properties", "required", "definitions", "$ref", "allOf"}


def _get_basic_template(openapi_schema: dict) -> dict:
    """For an OpenAPI parameter schema it prepares a template with basic fields"""
    template = {}
    for key, value in openapi_schema.items():  # get metadata fields
        if key in _SUPPORTED_OPENAPI_FIELDS or (key == "type" and value == "object"):
            continue

        template[f"#{key}"] = value

    return template


def _get_referred_model_name(openapi_schema: dict) -> Optional[str]:
    """In a parameter schema it finds a reference to another model schema. If it doesn't exist it returns None"""
    referred_path = openapi_schema.get("$ref")
    if not referred_path:
        schema_items = openapi_schema.get("items", {})  # items can be a list
        referred_path = schema_items.get("$ref") if isinstance(schema_items, dict) else None

    if not referred_path and "allOf" in openapi_schema:
        referred_path = openapi_schema["allOf"][0].get("$ref")

    if referred_path:
        return referred_path.rsplit("/", 1)[-1]
    return referred_path


def build_minimal_template(
    schema: Type[BaseModel],
    required_only: bool,
    pipeline_import_path: Optional[str] = None,
    add_descriptions: bool = False,
) -> dict:
    rec_flags = dict(required_only=required_only, add_descriptions=add_descriptions)
    json_schema = schema.schema()  # needed for descriptions

    template: dict = {}
    for name, field in schema.__fields__.items():

        if required_only and not field.required:
            continue

        if name == "pipeline" and pipeline_import_path:
            template[name] = pipeline_import_path
        elif isclass(field.type_) and issubclass(field.type_, BaseModel):
            # Contains a subschema in the nesting
            if isclass(field.outer_type_) and issubclass(field.outer_type_, BaseModel):
                template[name] = build_minimal_template(field.type_, **rec_flags)
                if "description" in json_schema["properties"][name]:
                    template[name]["<< description >>"] = json_schema["properties"][name]["description"]
            else:
                template[name] = {
                    "<< type >>": repr(field._type_display()),
                    "<< nested schema >>": str(field.type_),
                    "<< sub-template >>": build_minimal_template(field.type_, **rec_flags),
                }
        else:
            template[name] = _field_description(field, json_schema["properties"][name], add_descriptions)

    return template


def _field_description(field: ModelField, field_schema: dict, add_descriptions: bool) -> str:
    description = " // " + field_schema["description"] if "description" in field_schema and add_descriptions else ""
    field_type = repr(field._type_display())
    default = repr(field.default) + " : " if field.default else ""
    return f"<< {default}{field_type}{description} >>"



"""
Module implementing prediction pipeline
"""
import abc
from typing import Optional, List, Tuple

from pydantic import Field, validator

from eolearn.core import (
    FeatureType,
    EONode,
    SaveTask,
    LoadTask,
    OverwritePermission,
    EOWorkflow,
    MergeEOPatchesTask,
)

from ..core.pipeline import Pipeline
from ..tasks.prediction import ClassificationPredictionTask, RegressionPredictionTask
from ..utils.types import Feature
from ..utils.filter import get_patches_without_all_features


class BasePredictionPipeline(Pipeline, metaclass=abc.ABCMeta):
    """Pipeline to load a model and run prediction on EOPatches"""

    class Schema(Pipeline.Schema):
        input_folder_key: str = Field(
            description="The storage manager key pointing to the input folder of the model input data."
        )
        input_features: List[Feature] = Field(
            description=(
                "List of features of form `[(feature_type, feature_name)]` specifying which features are model input in"
                " the correct order"
            )
        )

        output_folder_key: str = Field(
            description="The storage manager key pointing to the output folder for the prediction pipeline."
        )

        prediction_mask_folder_key: Optional[str]
        prediction_mask_feature_name: Optional[str] = Field(
            description="Name of `MASK_TIMELESS` feature which defines which areas will be predicted"
        )

        @validator("prediction_mask_feature_name")
        def validate_filename(cls, v, values):  # pylint: disable=no-self-use, no-self-argument
            if values["prediction_mask_folder_key"] is None:
                assert v is None, "Both `prediction_mask_folder_key` and `prediction_mask_feature_name` must be given."
            return v

        model_folder_key: str = Field(
            description="The storage manager key pointing to the folder of the model used in the prediction pipeline."
        )
        model_filename: str = Field(description="A list of model filenames to be used for prediction")
        compress_level: int = Field(1, description="Level of compression used in saving EOPatches")

    @abc.abstractmethod
    def _get_output_features(self) -> List[Feature]:
        """Lists all features that are to be saved upon the pipeline completion"""

    @property
    def _is_mp_lock_needed(self) -> bool:
        """If a multiprocessing lock is needed when executing"""
        return not self.config.use_ray and self.config.workers > 1

    def filter_patch_list(self, patch_list: List[str]) -> List[str]:
        """EOPatches are filtered according to existence of specified output features"""

        filtered_patch_list = get_patches_without_all_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            self._get_output_features(),
        )

        return filtered_patch_list

    def build_workflow(self) -> EOWorkflow:
        """Workflow handling the prediction for eopatches.
        The workflow allows to add smoothing and custom thresholding to the predicted scores
        """
        preparation_node = self._get_data_preparation_node()
        predictions_node = self._get_prediction_node(preparation_node)
        saving_node = self._get_saving_node(predictions_node)

        return EOWorkflow.from_endnodes(saving_node)

    def _get_data_preparation_node(self) -> EONode:
        """Returns nodes containing for loading and preparing the data as well as the endpoint tasks"""
        features_load_node = EONode(
            LoadTask(
                self.storage.get_folder(self.config.input_folder_key, full_path=True),
                features=[FeatureType.BBOX, *self.config.input_features],
                config=self.sh_config,
            )
        )

        if not self.config.prediction_mask_folder_key:
            return features_load_node

        mask_load_node = EONode(
            LoadTask(
                self.storage.get_folder(self.config.prediction_mask_folder_key, full_path=True),
                features=[(FeatureType.MASK_TIMELESS, self.config.prediction_mask_feature_name)],
                config=self.sh_config,
            )
        )

        return EONode(MergeEOPatchesTask(), inputs=[features_load_node, mask_load_node])

    @abc.abstractmethod
    def _get_prediction_node(self, previous_node: EONode) -> EONode:
        """Returns nodes for using the model on the prepared data and the endpoints of the dependencies"""

    def _get_saving_node(self, previous_node: EONode) -> EONode:
        """Returns nodes for finalizing and saving features"""
        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key, full_path=True),
            features=self._get_output_features(),
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            config=self.sh_config,
            compress_level=self.config.compress_level,
        )

        return EONode(save_task, inputs=[previous_node])


class RegressionPredictionPipeline(BasePredictionPipeline):
    class Schema(BasePredictionPipeline.Schema):
        output_feature: str
        clip_predictions: Optional[Tuple[float, float]] = Field(
            description="Whether to clip values of predictions to specified interval"
        )

    def _get_output_features(self) -> List[Feature]:
        return [FeatureType.TIMESTAMP, FeatureType.BBOX, (FeatureType.DATA_TIMELESS, self.config.output_feature)]

    def _get_prediction_node(self, previous_node: EONode) -> EONode:
        prediction_task = RegressionPredictionTask(
            model_folder=self.storage.get_folder(self.config.model_folder_key, full_path=True),
            model_filename=self.config.model_filename,
            input_features=self.config.input_features,
            mask_feature=_optional_typed_feature(FeatureType.MASK_TIMELESS, self.config.prediction_mask_feature_name),
            output_feature=(FeatureType.DATA_TIMELESS, self.config.output_feature),
            mp_lock=self._is_mp_lock_needed,
            sh_config=self.sh_config,
            clip_predictions=self.config.clip_predictions,
        )
        return EONode(prediction_task, inputs=[previous_node])


class ClassificationPredictionPipeline(BasePredictionPipeline):
    class Schema(BasePredictionPipeline.Schema):
        output_feature: str
        output_probability_feature: Optional[str]

        label_encoder_filename: Optional[str] = Field(
            description=(
                "Whether the predictions need to be decoded. The label encoder should be in the same model folder."
            )
        )

    def _get_output_features(self) -> List[Feature]:
        out = [FeatureType.TIMESTAMP, FeatureType.BBOX, (FeatureType.MASK_TIMELESS, self.config.output_feature)]
        if self.config.output_probability_feature:
            out.append((FeatureType.DATA_TIMELESS, self.config.output_probability_feature))
        return out

    def _get_prediction_node(self, previous_node: EONode) -> EONode:
        prediction_task = ClassificationPredictionTask(
            model_folder=self.storage.get_folder(self.config.model_folder_key, full_path=True),
            model_filename=self.config.model_filename,
            input_features=self.config.input_features,
            mask_feature=_optional_typed_feature(FeatureType.MASK_TIMELESS, self.config.prediction_mask_feature_name),
            output_feature=(FeatureType.MASK_TIMELESS, self.config.output_feature),
            output_probability_feature=_optional_typed_feature(
                FeatureType.DATA_TIMELESS, self.config.output_probability_feature
            ),
            mp_lock=self._is_mp_lock_needed,
            sh_config=self.sh_config,
            label_encoder_filename=self.config.label_encoder_filename,
        )
        return EONode(prediction_task, inputs=[previous_node])


def _optional_typed_feature(ftype: FeatureType, fname: Optional[str]) -> Optional[Feature]:
    """Constructs a typed feature if possible"""
    return (ftype, fname) if fname is not None else None



"""
Module implementing utilities for unit testing pipeline results
"""
import os
import json
import functools
from typing import Tuple, Optional, Dict, List

import fs
import pandas as pd
import shapely.ops
import numpy as np
import rasterio
from deepdiff import DeepDiff

from eolearn.core import EOPatch, FeatureType

from ..core.config import Config
from ..core.pipeline import Pipeline
from .meta import load_pipeline


class ContentTester:
    """Utility used for testing pipeline results

    Results of a pipeline is usually a folder containing multiple EOPatches, each containing various features. This
    utility aggregates all folder content into some basic statistics.

    * Every time you initialize this class statistics will be calculated
    * Statistics can be saved into a JSON file (it's human readable)
    * Statistics can be compared compared with the one saved in a file from the previous run.

    If statistics match there is a good chance that the pipeline produced exactly the same results as before. Otherwise
    this utility will let you know which statistics does not match
    """

    def __init__(self, filesystem: fs.base.FS, main_folder: str, decimals: int = 5):
        """
        :param filesystem: A filesystem containing project data
        :param main_folder: A folder path on the filesystem where results are saved
        :param decimals: Number of decimals to which values will be rounded
        """
        self.filesystem = filesystem
        self.main_folder = main_folder
        self.decimals = decimals

        if not filesystem.isdir(self.main_folder):
            raise ValueError(f"Folder {self.main_folder} does not exist on filesystem {self.filesystem}")

        self.stats = self._calculate_stats()

    def compare(self, filename: str) -> DeepDiff:
        """Compares statistics of given folder content with statistics saved in a given file

        :param filename: A JSON filename (with file path) where expected statistics is saved
        :return: A dictionary report about differences between expected and actual content
        """
        with open(filename, "r") as file:
            expected_stats = json.load(file)

        return DeepDiff(expected_stats, self.stats)

    def save(self, filename: str):
        """Saves statistics of given folder content into a JSON file

        :param filename: A JSON filename (with file path) where statistics should be saved
        """
        with open(filename, "w") as file:
            json.dump(self.stats, file, indent=2, sort_keys=True)

    def _calculate_stats(self, folder: Optional[str] = None) -> Dict[str, object]:
        """Calculates statistics of given folder and it's content"""
        stats = {}
        if folder is None:
            folder = self.main_folder

        for content in self.filesystem.listdir(folder):
            content_path = fs.path.combine(folder, content)

            if self.filesystem.isdir(content_path):
                eopatch = EOPatch.load(content_path, filesystem=self.filesystem)

                if eopatch.bbox:
                    stats[content] = self._calculate_eopatch_stats(eopatch)
                else:  # Probably it is not an EOPatch folder
                    stats[content] = self._calculate_stats(folder=content_path)
            elif content_path.endswith("tiff"):
                stats[content] = self._calculate_tiff_stats(content_path)
            elif content_path.endswith(".npy"):
                stats[content] = self._calculate_numpy_stats(content_path)
            elif content_path.endswith(".aux.xml"):
                pass
            else:
                stats[content] = None

        return stats

    def _calculate_eopatch_stats(self, eopatch: EOPatch) -> Dict[str, object]:
        """Calculates statistics of given EOPatch and it's content"""
        stats = {}

        for feature_type, feature_set in eopatch.get_features().items():
            feature_type_name = feature_type.value

            if feature_type is FeatureType.BBOX:
                stats[feature_type_name] = repr(eopatch.bbox)

            elif feature_type is FeatureType.TIMESTAMP:
                stats[feature_type_name] = [time.isoformat() for time in eopatch.timestamp]

            else:
                stats[feature_type_name] = {}

                if feature_type.is_raster():
                    calculation_method = self._calculate_raster_stats
                elif feature_type.is_vector():
                    calculation_method = self._calculate_vector_stats
                else:  # Only FeatureType.META_INFO remains
                    calculation_method = str

                for feature_name in feature_set:
                    stats[feature_type_name][feature_name] = calculation_method(eopatch[feature_type][feature_name])

        return stats

    def _calculate_raster_stats(self, raster: np.ndarray) -> List[object]:
        """Calculates statistics over a raster numpy array"""
        stats = [*raster.shape, str(raster.dtype)]
        if raster.dtype == object or raster.dtype.kind == "U":
            return stats

        stats.extend([np.count_nonzero(np.isnan(raster)), np.count_nonzero(np.isfinite(raster))])

        finite_values = raster[np.isfinite(raster)]

        if finite_values.size:
            if finite_values.dtype == bool:
                finite_values = finite_values.astype(np.uint8)

            for operation in [np.min, np.max, np.mean, np.median]:
                stats.append(round(float(operation(finite_values)), self.decimals))

            stats.extend(map(int, np.histogram(finite_values, bins=8)[0]))

            np.random.seed(0)
            randomly_chosen_values = np.random.choice(finite_values, 8)
            stats.extend(map(lambda value: round(float(value), self.decimals), randomly_chosen_values))

        return stats

    def _calculate_tiff_stats(self, tiff_filename: str) -> List[object]:
        """Calculates statistics over a .tiff image"""
        with self.filesystem.openbin(tiff_filename, "r") as file_handle:
            with rasterio.open(file_handle) as tiff:
                raster = tiff.dataset_mask()

        return self._calculate_raster_stats(raster)

    def _calculate_numpy_stats(self, numpy_filename: str) -> List[object]:
        """Calculates statistics over a .npy file containing a numpy array"""
        with self.filesystem.openbin(numpy_filename, "r") as file_handle:
            raster = np.load(file_handle, allow_pickle=True)

        return self._calculate_raster_stats(raster)

    def _calculate_vector_stats(self, dataframe: pd.DataFrame) -> List[object]:
        """Calculates statistics over a vector GeoDataFrame"""  # TODO: add more statistical properites
        rounder = functools.partial(_round_point_coords, decimals=self.decimals)
        dataframe["geometry"] = dataframe["geometry"].apply(lambda geometry: shapely.ops.transform(rounder, geometry))

        stats = list(dataframe) + [len(dataframe.index), str(dataframe.crs)]

        if len(dataframe.index):
            stats.extend(list(map(str, dataframe.iloc[0])))

        return stats


def check_pipeline_logs(pipeline: Pipeline):
    """A utility function which checks pipeline logs and makes sure there are no failed executions"""
    if not pipeline.config.logging.save_logs:
        raise ValueError("Pipeline did not save logs, this test would be useless")

    logs_folder = pipeline.logging_manager.get_pipeline_logs_folder(pipeline.current_execution_name)

    for filename in ["failed.json", "finished.json", "pipeline-report.json", "pipeline.log"]:
        path = fs.path.combine(logs_folder, filename)
        assert pipeline.storage.filesystem.isfile(path), f"File {path} is missing"

    logs_folder = pipeline.logging_manager.get_pipeline_logs_folder(pipeline.current_execution_name, full_path=True)
    failed_filename = fs.path.combine(logs_folder, "failed.json")
    assert not pipeline.eopatch_manager.load_eopatch_filenames(
        failed_filename
    ), f"Some executions failed, check {logs_folder}"

    finished_filename = os.path.join(logs_folder, "finished.json")
    assert pipeline.eopatch_manager.load_eopatch_filenames(finished_filename), "No executions finished"


def run_and_test_pipeline(
    config_folder: str,
    config_name: str,
    stats_folder: str,
    stats_name: str,
    folder_key: str,
    reset_folder: bool = True,
    save_new_stats: bool = False,
):
    """A default way of testing a pipeline

    :param config_folder: A path to folder containing the config file
    :param config_name: A config filename
    :param stats_folder: A path to folder containing the file with expected result stats
    :param stats_name: A filename of expected results stats
    :param folder_key: Type of the folder containing results of the pipeline
    :param reset_folder: If True it will delete content of the folder with results before running the pipeline
    :param save_new_stats: If True then new file with expected result stats will be saved (potentially overwriting the
        old one. Otherwise the old one will be used to compare stats.
    """
    config_filename = os.path.join(config_folder, config_name)
    expected_stats_file = os.path.join(stats_folder, stats_name)

    config = Config.from_path(config_filename)
    pipeline = load_pipeline(config)

    folder = pipeline.storage.get_folder(folder_key)
    filesystem = pipeline.storage.filesystem

    if reset_folder:
        filesystem.removetree(folder)
    pipeline.run()

    check_pipeline_logs(pipeline)

    tester = ContentTester(filesystem, folder)

    if save_new_stats:
        tester.save(expected_stats_file)

    stats_difference = tester.compare(expected_stats_file)
    if stats_difference:
        raise AssertionError(f"Expected and obtained stats differ: {stats_difference}")


def _round_point_coords(x: float, y: float, decimals: int) -> Tuple[float, float]:
    """Rounds coordinates of a point"""
    return round(x, decimals), round(y, decimals)



"""
This module handles everything regarding storage of the data
"""
from typing import Optional, Dict, List
from io import StringIO

import fs
from pydantic import Field

from sentinelhub import SHConfig
from eolearn.core.utils.fs import get_filesystem, is_s3_path, get_aws_credentials

from .base import EOGrowObject
from .schemas import ManagerSchema


class StorageManager(EOGrowObject):

    PRESET_FOLDERS: Dict[str, str] = {"logs": "logs", "input_data": "input-data", "cache": "cache"}

    class Schema(ManagerSchema):
        project_folder: str = Field(
            description=(
                "The root project folder. Can be either local or on AWS S3 Bucket."
                "If on AWS, the path must be prefixed with s3://."
            ),
        )
        aws_profile: Optional[str] = Field(
            description="The AWS profile with credentials needed to access the S3 bucket"
        )
        structure: Dict[str, str] = Field(
            default_factory=dict, description="A flat key: value store mapping each key to a path in the project."
        )

    def __init__(self, config: dict):
        super().__init__(config)

        for folder_key, folder_path in self.PRESET_FOLDERS.items():
            if folder_key not in self.config.structure:
                self.config.structure[folder_key] = folder_path

        self.sh_config = SHConfig(hide_credentials=True)
        if self.is_on_aws() and self.config.aws_profile:
            self.sh_config = get_aws_credentials(aws_profile=self.config.aws_profile, config=self.sh_config)

        self.filesystem = get_filesystem(self.config.project_folder, create=True, config=self.sh_config)

    def get_folder(self, key: str, full_path: bool = False) -> str:
        """Returns the path  associated with a key in the structure config."""
        folder_path = self.config.structure[key]
        self.filesystem.makedirs(folder_path, recreate=True)

        if full_path:
            return fs.path.combine(self.config.project_folder, folder_path)
        return folder_path

    def get_logs_folder(self, full_path: bool = False) -> str:
        """Method for obtaining the logs folder. Will store logs to the current folder.
        Temporary solution until the logging to AWS is handled properly
        """
        return self.get_folder("logs", full_path=full_path)

    def get_cache_folder(self, full_path: bool = False) -> str:
        """Returns the path associated with the cache key."""
        return self.get_folder("cache", full_path=full_path)

    def get_input_data_folder(self, full_path: bool = False) -> str:
        """Returns the path associated with the input_data key."""
        return self.get_folder("input_data", full_path=full_path)

    def is_on_aws(self) -> bool:
        """Returns True if the project_folder is on S3, False  otherwise."""
        return is_s3_path(self.config.project_folder)

    def show_folder_structure(
        self, show_files: bool = False, return_str: bool = False, exclude: Optional[List[str]] = None
    ) -> Optional[str]:
        """Shows how folder structure looks like at the moment. It will show all folders except EOPatch folders and
        EOExecution report folders

        :param show_files: If  `True` it will show also files inside the folders. Note that the number of files may be
            huge. By default this is set to `False`.
        :param return_str: If `True` it will return folder structure as a string. If `False` it will just print the
            visualization to stdout.
        :param exclude: A list of grep folder paths  to exclude from the structure return.
            Defaults to ['eopatch*', 'eoexecution-report*']
        :return: Depending on return_str it will either return a string or None
        """
        if exclude is None:
            exclude = ["eopatch*", "eoexecution-report*"]

        file_filter = None if show_files else [""]
        io_object = StringIO() if return_str else None
        self.filesystem.tree(max_levels=10, with_color=True, exclude=exclude, filter=file_filter, file=io_object)
        if io_object:
            io_object.seek(0)
            return io_object.getvalue()
        return None



"""
A module containing general utilities that haven't been sorted in any other module
"""
import datetime as dt
from enum import Enum

from aenum import MultiValueEnum

from sentinelhub import DataCollection
from sentinelhub.data_collections import DataCollectionDefinition


def jsonify(param: object) -> str:
    """Transforms an object into a normal string."""
    if isinstance(param, (dt.datetime, dt.date)):
        return param.isoformat()

    if isinstance(param, DataCollectionDefinition):
        return DataCollection(param).name
    if isinstance(param, DataCollection):
        return param.name

    if isinstance(param, MultiValueEnum):
        return param.values[0]

    if isinstance(param, Enum):
        return param.value

    raise TypeError(f"Object of type {type(param)} is not yet supported in jsonify utility function")



"""
Pipelines for testing
"""
import logging
from typing import List, Tuple

from ..core.config import recursive_config_join
from ..core.pipeline import Pipeline

LOGGER = logging.getLogger(__name__)


class TestPipeline(Pipeline):
    """Pipeline that just tests if all managers works correctly. It can be used to check if area manager creates a
    correct grid.
    """

    class Schema(Pipeline.Schema):
        class Config:
            extra = "allow"

    _DEFAULT_CONFIG_PARAMS = {
        "pipeline": "eogrow.pipelines.testing.TestPipeline",
        "eopatch": {"manager": "eogrow.eopatches.EOPatchManager"},
        "logging": {"manager": "eogrow.logging.LoggingManager", "show_logs": True},
    }

    def __init__(self, config: dict):
        config = recursive_config_join(config, self._DEFAULT_CONFIG_PARAMS)
        super().__init__(config)

    def run_procedure(self) -> Tuple[List, List]:
        """Performs basic tests of managers"""
        if self.storage.filesystem.exists("/"):
            LOGGER.info("Project folder %s exists", self.storage.config.project_folder)
        else:
            LOGGER.info("Project folder %s does not exist", self.storage.config.project_folder)

        self.area_manager.get_area_dataframe()
        self.area_manager.get_area_geometry()
        grid = self.area_manager.get_grid()
        grid_size = self.area_manager.get_grid_size()
        LOGGER.info("Grid has %d EOPatches and is split over %d CRS zones", grid_size, len(grid))

        eopatches = self.eopatch_manager.get_eopatch_filenames()
        LOGGER.info("The first EOPatch has a name %s", eopatches[0])

        return [], []



"""
Tasks used to transform Sentinel Hub Batch results into EOPatches
"""
import json
import concurrent.futures
from typing import Optional, List

import fs
import numpy as np
from sentinelhub import parse_time, SHConfig
from eolearn.core import EOPatch, EOTask
from eolearn.core.utils.fs import get_base_filesystem_and_path

from ..utils.meta import import_object
from ..utils.types import Feature


class LoadUserDataTask(EOTask):
    """Task that loads and adds timestamps and tile names to an EOPatch"""

    def __init__(
        self,
        path: str,
        userdata_feature_name: Optional[str] = None,
        userdata_timestamp_reader: Optional[str] = None,
        config: Optional[SHConfig] = None,
    ):
        """
        :param path: Path to folder containing the tiles
        :param userdata_feature_name: A name of a META_INFO feature in which userdata.json content could be stored
        :param userdata_timestamp_reader: A reference to a Python function or a Python code that collects timestamps
            from loaded userdata.json
        :param config: A configuration object with AWS credentials
        """
        self.path = path
        self.userdata_feature_name = userdata_feature_name
        self.userdata_timestamp_reader = userdata_timestamp_reader
        self.config = config

    def _load_userdata_file(self, folder: str, filename: str = "userdata.json") -> dict:
        """Loads a content of a JSON file"""
        filesystem, relative_path = get_base_filesystem_and_path(self.path, config=self.config)
        full_path = fs.path.join(relative_path, folder, filename)

        userdata_text = filesystem.readtext(full_path, encoding="utf-8")
        return json.loads(userdata_text)

    @staticmethod
    def _parse_timestamps(userdata: dict, userdata_timestamp_reader: str) -> list:
        """Parses timestamps from userdata dictionary"""
        try:
            reader = import_object(userdata_timestamp_reader)
            time_strings = reader(userdata)
        except (ImportError, ValueError):
            time_strings = eval(userdata_timestamp_reader)  # pylint: disable=eval-used

        return [parse_time(time_string, force_datetime=True, ignoretz=True) for time_string in time_strings]

    def execute(self, eopatch: Optional[EOPatch], *, folder: str) -> EOPatch:
        """Adds metadata to the given EOPatch

        :param eopatch: Name of the eopatch to process
        :param folder: Folder in which userdata.json is stored
        """
        eopatch = eopatch or EOPatch()

        userdata = self._load_userdata_file(folder)

        if self.userdata_feature_name:
            eopatch.meta_info[self.userdata_feature_name] = userdata

        if self.userdata_timestamp_reader:
            eopatch.timestamp = self._parse_timestamps(userdata, self.userdata_timestamp_reader)

        return eopatch


class FixImportedTimeDependentFeatureTask(EOTask):
    """Fixes a time-dependent feature that has been imported as a timeless feature from batch results.

    It performs the following:
    - rotates bands axis to time axis,
    - reverses order of timestamps and feature
    - potentially removes redundant timeframes according to timestamps. This is necessary because in case there were
      no available acquisitions batch job still had to save a single time frame with dummy values.
    """

    def __init__(self, input_feature: Feature, output_feature: Feature):
        self.input_feature = input_feature
        self.output_feature = output_feature

    def execute(self, eopatch: EOPatch) -> EOPatch:
        """Fixes a feature in the given EOPatch"""
        data = eopatch[self.input_feature]
        del eopatch[self.input_feature]

        data = data[np.newaxis, ...]
        data = np.swapaxes(data, 0, -1)

        if eopatch.timestamp:
            timeframe_num = len(eopatch.timestamp)
            if data.shape[0] != timeframe_num:  # Handling a case where data would contain some empty timeframes
                data = data[:timeframe_num, ...]

            order_mask = np.argsort(eopatch.timestamp)
            is_strictly_increasing = (np.diff(order_mask) > 0).all()
            if not is_strictly_increasing:
                eopatch.timestamp = sorted(eopatch.timestamp)
                data = data[order_mask]

        eopatch[self.output_feature] = data
        return eopatch


class DeleteFilesTask(EOTask):
    """Delete files"""

    def __init__(self, path: str, filenames: List[str], config: Optional[SHConfig] = None):
        """
        :param path: Path to folder containing the files to be deleted
        :type path: str
        :param filenames: A list of filenames to delete
        :type filenames: list(str)
        :param config: A configuration object with AWS credentials
        :type config: SHConfig
        """
        self.path = path
        self.filenames = filenames
        self.config = config

    def execute(self, *_, folder: str) -> None:
        """Execute method to delete files relative to the specified tile

        :param folder: A folder containing files
        """
        filesystem, relative_path = get_base_filesystem_and_path(self.path, config=self.config)

        file_paths = [fs.path.join(relative_path, folder, filename) for filename in self.filenames]
        with concurrent.futures.ThreadPoolExecutor() as executor:
            # The following is intentionally wrapped in a list in order to get back potential exceptions
            list(executor.map(filesystem.remove, file_paths))



import os
from setuptools import setup, find_packages


def get_long_description():
    this_directory = os.path.abspath(os.path.dirname(__file__))

    with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
        long_description = f.read()

    # Removes lines with an image from description
    return "\n".join(line for line in long_description.split("\n") if not line.strip().startswith("!["))


def parse_requirements(file):
    required_packages = []
    with open(os.path.join(os.path.dirname(__file__), file)) as req_file:
        for line in req_file:
            if "/" not in line:
                required_packages.append(line.strip())
    return required_packages


def get_version():
    for line in open(os.path.join(os.path.dirname(__file__), "eogrow", "__init__.py")):
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"').strip("'")
    return version


setup(
    name="eo-grow",
    python_requires=">=3.8",
    version=get_version(),
    description="Earth observation framework for scaled-up processing in Python",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/sentinel-hub/eo-grow",
    author="Sinergise EO research team",
    author_email="eoresearch@sinergise.com",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=parse_requirements("requirements.txt"),
    extras_require={"DEV": parse_requirements("requirements-dev.txt")},
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "eogrow=eogrow.cli:EOGrowCli.main",
            "eogrow-ray=eogrow.cli:EOGrowCli.ray",
            "eogrow-template=eogrow.cli:EOGrowCli.make_template",
            "eogrow-validate=eogrow.cli:EOGrowCli.validate_config",
            "eogrow-test=eogrow.cli:EOGrowCli.run_test_pipeline",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
)



"""
Module implementing pipelines for downloading data
"""
import datetime as dt
import logging
import abc
from typing import Optional, List, Tuple, Union

import numpy as np
from pydantic import Field
from sentinelhub import DataCollection, read_data, MimeType, Band, Unit
from eolearn.core import FeatureType, OverwritePermission, EOWorkflow, EONode, SaveTask
from eolearn.io import SentinelHubInputTask, SentinelHubDemTask, SentinelHubEvalscriptTask
from eolearn.features import LinearFunctionTask

from ..core.config import Config
from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..utils.validators import (
    field_validator,
    optional_field_validator,
    parse_time_period,
    parse_data_collection,
    validate_mosaicking_order,
    validate_resampling,
)
from ..utils.types import TimePeriod, Path, Feature
from ..utils.filter import get_patches_without_all_features

LOGGER = logging.getLogger(__name__)


class RescaleSchema(BaseSchema):
    rescale_factor: float = Field(1, description="Amount by which the selected features are multiplied")
    dtype: Optional[str] = Field(description="The output dtype of data")
    features_to_rescale: List[Feature]


class PostprocessingRescale(BaseSchema):
    rescale_schemas: List[RescaleSchema] = Field(
        default_factory=list, description="Specify different ways to rescale features"
    )


class BaseDownloadPipeline(Pipeline, metaclass=abc.ABCMeta):
    """Base pipeline for downloading satellite data"""

    class Schema(Pipeline.Schema):
        output_folder_key: str = Field(
            description="Storage manager key pointing to the path where downloaded EOPatches will be saved."
        )

        compress_level: int = Field(1, description="Level of compression used in saving EOPatches")
        threads_per_worker: Optional[int] = Field(
            description=(
                "Maximum number of parallel threads used during download by each worker. If set to None it will use "
                "5 * N threads, where N is the number of CPUs on the machine"
            ),
        )

        postprocessing: Optional[PostprocessingRescale] = Field(description="Parameters used in post-processing tasks")

    def __init__(self, config: Config):
        super().__init__(config)
        self.download_node_uid = None

    def filter_patch_list(self, patch_list: List[str]) -> List[str]:
        """EOPatches are filtered according to existence of specified output features"""

        filtered_patch_list = get_patches_without_all_features(
            self.storage.filesystem,
            self.storage.get_folder(self.config.output_folder_key),
            patch_list,
            self._get_output_features(),
        )

        return filtered_patch_list

    @abc.abstractmethod
    def _get_output_features(self) -> List[Feature]:
        """Lists all features that are to be saved upon the pipeline completion"""

    @abc.abstractmethod
    def _get_download_node(self) -> EONode:
        """Provides node for downloading data."""

    @staticmethod
    def get_postprocessing_node(postprocessing_config: Config, previous_node: EONode) -> EONode:
        """Provides node that applies postprocessing to data after download is complete"""
        node = previous_node
        for rescale_config in postprocessing_config.rescale_schemas:
            node = EONode(
                LinearFunctionTask(
                    rescale_config.features_to_rescale, slope=rescale_config.rescale_factor, dtype=rescale_config.dtype
                ),
                inputs=[node],
            )

        return node

    def build_workflow(self) -> EOWorkflow:
        """Method that builds a workflow"""
        download_node = self._get_download_node()
        self.download_node_uid = download_node.uid

        postprocessing_node = None
        if self.config.postprocessing:
            postprocessing_node = self.get_postprocessing_node(self.config.postprocessing, download_node)

        end_node = EONode(
            SaveTask(
                self.storage.get_folder(self.config.output_folder_key, full_path=True),
                features=self._get_output_features(),
                compress_level=self.config.compress_level,
                overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
                config=self.sh_config,
            ),
            inputs=[postprocessing_node or download_node],
        )

        return EOWorkflow.from_endnodes(end_node)

    def get_execution_arguments(self, workflow: EOWorkflow) -> List[dict]:
        """Adds required bbox and time_interval parameters for input task to the base execution arguments

        :param workflow: EOWorkflow used to download images
        """
        exec_args = super().get_execution_arguments(workflow)

        download_node = workflow.get_node_with_uid(self.download_node_uid)
        if download_node is None:
            return exec_args

        bbox_list = self.eopatch_manager.get_bboxes(eopatch_list=self.patch_list)

        for index, bbox in enumerate(bbox_list):
            exec_args[index][download_node] = {"bbox": bbox}
            if "time_period" in self.config:
                exec_args[index][download_node]["time_interval"] = self.config.time_period

        return exec_args


class CommonDownloadFields(BaseSchema):
    data_collection: DataCollection = Field(description="Data collection from which data will be downloaded.")
    _validate_data_collection = field_validator("data_collection", parse_data_collection, pre=True)

    resolution: float = Field(description="Resolution of downloaded data in meters")

    maxcc: Optional[float] = Field(ge=0, le=1)

    resampling_type: Optional[str] = Field(
        description="A type of downsampling and upsampling used by Sentinel Hub service. Default is NEAREST"
    )
    _validate_resampling_type = optional_field_validator("resampling_type", validate_resampling)


class TimeDependantFields(BaseSchema):
    time_period: TimePeriod
    _validate_time_period = field_validator("time_period", parse_time_period, pre=True)

    time_difference: Optional[float] = Field(description="Time difference in minutes between consecutive time frames")

    mosaicking_order: Optional[str] = Field(
        description="The mosaicking order used by Sentinel Hub service. Default is mostRecent"
    )
    _validate_mosaicking_order = optional_field_validator("mosaicking_order", validate_mosaicking_order)


class DownloadPipeline(BaseDownloadPipeline):
    """Pipeline to download data via SentinelHubInputTask"""

    class Schema(BaseDownloadPipeline.Schema, CommonDownloadFields, TimeDependantFields):
        bands_feature_name: str = Field(description="Name of a feature in which bands will be saved")
        bands: Optional[List[str]] = Field(description="Names of bands to download")
        additional_data: List[Feature] = Field(default_factory=list, description="Additional data to download")
        use_dn: bool = Field(
            False, description="Whether to save bands as float32 reflectance (default), or int16 digital numbers."
        )

    def _get_output_features(self) -> List[Feature]:
        base_features = [(FeatureType.DATA, self.config.bands_feature_name), FeatureType.BBOX, FeatureType.TIMESTAMP]
        return base_features + self.config.additional_data

    def _get_download_node(self) -> EONode:
        time_diff = None if self.config.time_difference is None else dt.timedelta(minutes=self.config.time_difference)
        bands_dtype = np.uint16 if self.config.use_dn else np.float32
        data_collection = self.config.data_collection

        if data_collection.is_byoc:
            data_collection = self.config.data_collection.define_from(
                f"{self.config.data_collection.name}_WITH_BANDS",
                bands=[Band(name=band, units=(Unit.DN,), output_types=(bands_dtype,)) for band in self.config.bands],
                metabands=[(Band(name="dataMask", units=(Unit.DN,), output_types=(bool,)))],
            )

        download_task = SentinelHubInputTask(
            bands_feature=(FeatureType.DATA, self.config.bands_feature_name),
            bands=self.config.bands,
            resolution=self.config.resolution,
            maxcc=self.config.maxcc,
            time_difference=time_diff,
            data_collection=data_collection,
            max_threads=self.config.threads_per_worker,
            additional_data=self.config.additional_data,
            bands_dtype=bands_dtype,
            config=self.sh_config,
            mosaicking_order=self.config.mosaicking_order,
            aux_request_args=_get_aux_request_args(self.config.resampling_type),
        )
        return EONode(download_task)


class DownloadEvalscriptPipeline(BaseDownloadPipeline):
    """Pipeline to download through an evalscript"""

    class Schema(BaseDownloadPipeline.Schema, CommonDownloadFields, TimeDependantFields):
        features: List[Union[Feature, Tuple[FeatureType, str, str]]] = Field(
            description="Features to construct from the evalscript, with the options to provide new names"
        )
        evalscript_path: Path

    def _get_output_features(self) -> List[Feature]:
        return self.config.features + [FeatureType.BBOX, FeatureType.TIMESTAMP]

    def _get_download_node(self) -> EONode:
        evalscript = read_data(self.config.evalscript_path, data_format=MimeType.TXT)
        time_diff = None if self.config.time_difference is None else dt.timedelta(minutes=self.config.time_difference)

        download_task = SentinelHubEvalscriptTask(
            features=self.config.features,
            evalscript=evalscript,
            data_collection=self.config.data_collection,
            resolution=self.config.resolution,
            maxcc=self.config.maxcc,
            time_difference=time_diff,
            max_threads=self.config.threads_per_worker,
            config=self.sh_config,
            mosaicking_order=self.config.mosaicking_order,
            aux_request_args=_get_aux_request_args(self.config.resampling_type),
        )
        return EONode(download_task)


class DownloadTimelessPipeline(BaseDownloadPipeline):
    """Pipeline to download timeless data"""

    class Schema(BaseDownloadPipeline.Schema, CommonDownloadFields):
        feature_name: str = Field(description="Name of the resulting feature")

    def _get_output_features(self) -> List[Feature]:
        return [(FeatureType.DATA_TIMELESS, self.config.feature_name), FeatureType.BBOX]

    def _get_download_node(self) -> EONode:

        download_task = SentinelHubDemTask(
            feature=(FeatureType.DATA_TIMELESS, self.config.feature_name),
            data_collection=self.config.data_collection,
            resolution=self.config.resolution,
            maxcc=self.config.maxcc,
            max_threads=self.config.threads_per_worker,
            config=self.sh_config,
            aux_request_args=_get_aux_request_args(self.config.resampling_type),
        )
        return EONode(download_task)


def _get_aux_request_args(resampling: Optional[str]) -> Optional[dict]:
    if resampling is not None:
        return {"processing": {"downsampling": resampling, "upsampling": resampling}}
    return None
